import inspect
import os
import pathlib
import shutil
from tempfile import TemporaryDirectory
from typing import Any

import kim_edn

import kimvv


def get_default_args(func):
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


class KIMVVTestDriver:
    @classmethod
    def get_kimspec(cls):
        mypath = pathlib.Path(__file__).parent.resolve()
        myname = cls.__name__
        return kim_edn.load(os.path.join(mypath, myname, "kimspec.edn"))

    def _resolve_dependencies(self, material, **kwargs):
        """
        defaults to equilibrium but can be defined within
        TestDriver for driver specific needs
        """
        # updates kwargs if needed and returns
        if isinstance(self, kimvv.EquilibriumCrystalStructure):
            return material, kwargs
        else:
            print("Resolving dependencies...")
            ecs_test = kimvv.EquilibriumCrystalStructure(self._calc)
            ecs_results = ecs_test(material)
            for result in ecs_results:
                if result["property-id"].endswith("crystal-structure-npt"):
                    material_relaxed = result
                    break
            return material_relaxed, kwargs

    @classmethod
    def printdoc(cls):
        print("\nDescription of method and non-structure arguments:\n")
        print(cls._calculate.__doc__)
        print("\nDefaults:")
        print(get_default_args(cls._calculate))
        print()


# new call decorator
def override_call_method(cls):
    def __call__(self, material: Any = None, **kwargs):
        """
        Taken from kim-tools with added dependency functionality
        Main operation of a Test Driver:

            * Call :func:`~KIMTestDriver._init_output_dir`
            * Run :func:`~KIMTestDriver._setup` (the base class provides a barebones
              version, derived classes may override)
            * Call :func:`~KIMTestDriver._calculate` (implemented by each individual
              Test Driver)
            * Call :func:`~KIMTestDriver._archive_aux_files`

        Args:
            material:
                Placeholder object for arguments describing the material to run
                the Test Driver on

        Returns:
            The property instances calculated during the current run
        """
        # count how many instances we had before we started
        previous_properties_end = len(self.property_instances)

        # Set up the output directory
        self._init_output_dir()

        with TemporaryDirectory() as d:
            original_output_bak = shutil.move("output", d)
            try:
                # resolve dependencies
                # since input to calculate may depend on output, return kwargs
                material_relaxed, kwargs = self._resolve_dependencies(
                    material, **kwargs
                )
            finally:
                # Dependencies wrote an output directory
                if os.path.exists("output"):
                    i = 0
                    while os.path.exists(f"output.dependency.{i}"):
                        i += 1
                    dep_output_bak = f"output.dependency.{i}"
                    msg = (
                        "Backing up 'output' directory written by dependency to "
                        f"{dep_output_bak}"
                    )
                    print(msg)
                    os.rename("output", dep_output_bak)
                # restore original output
                shutil.move(original_output_bak, ".")

        try:
            # _setup is likely overridden by an derived class
            self._setup(material_relaxed, **kwargs)

            # implemented by each individual Test Driver
            self._calculate(**kwargs)
        finally:
            # Postprocess output directory for this invocation
            self._archive_aux_files()

        # The current invocation returns a Python list of dictionaries containing all
        # properties computed during this run
        return self.property_instances[previous_properties_end:]

    cls.__call__ = __call__
    return cls



"""
Performs a symmetry-constrained minimization of the crystal
"""

from kim_tools import (
    SingleCrystalTestDriver,
    minimize_wrapper,
    get_stoich_reduced_list_from_prototype,
)
from ase.filters import FrechetCellFilter, UnitCellFilter
from ase.optimize.optimize import Optimizer
from ase.optimize.lbfgs import LBFGSLineSearch
from typing import Dict


class TestDriver(SingleCrystalTestDriver):
    def _calculate(
        self,
        fmax: float = 1e-5,
        steps: int = 200,
        algorithm: Optimizer = LBFGSLineSearch,
        cell_filter: UnitCellFilter = FrechetCellFilter,
        opt_kwargs: Dict = {},
        flt_kwargs: Dict = {},
        **kwargs
    ) -> None:
        """
        Performs a symmetry-constrained minimization of the crystal

        Args:
            fmax:
                Maximum force component tolerance
            steps:
                Maximum number of steps
        """
        atoms = self._get_atoms()
        num_atoms = len(atoms)
        prototype_label = self._get_nominal_crystal_structure_npt()["prototype-label"][
            "source-value"
        ]
        stoichiometry = get_stoich_reduced_list_from_prototype(prototype_label)
        stoichiometric_species = self._get_nominal_crystal_structure_npt()[
            "stoichiometric-species"
        ]["source-value"]
        num_atoms_in_formula = sum(stoichiometry)
        minimization_succeeded = minimize_wrapper(
            atoms=atoms,
            fmax=fmax,
            steps=steps,
            logfile="-",
            algorithm=algorithm,
            cell_filter=cell_filter,
            fix_symmetry=True,
            opt_kwargs=opt_kwargs,
            flt_kwargs=flt_kwargs,
        )
        potential_energy = atoms.get_potential_energy()
        potential_energy_per_atom = potential_energy / num_atoms
        potential_energy_per_formula = potential_energy_per_atom * num_atoms_in_formula
        isolated_energy_per_formula = sum(
            [
                num_per_formula
                * self.get_isolated_energy_per_atom(species)
                for num_per_formula, species in zip(
                    stoichiometry, stoichiometric_species
                )
            ]
        )
        binding_potential_energy_per_formula = (
            potential_energy_per_formula - isolated_energy_per_formula
        )
        binding_potential_energy_per_atom = (
            binding_potential_energy_per_formula / num_atoms_in_formula
        )
        self._update_nominal_parameter_values(atoms)
        if minimization_succeeded:
            disclaimer = None
        else:
            disclaimer = (
                "The forces and stresses failed to converge "
                "to the requested tolerance"
            )
            print(
                "\nThe minimization failed to converge. "
                "See kim-tools.log for more info"
            )

        self._add_property_instance_and_common_crystal_genome_keys(
            "crystal-structure-npt",
            write_stress=True,
            write_temp=True,
            disclaimer=disclaimer,
        )
        self._add_property_instance_and_common_crystal_genome_keys(
            "binding-energy-crystal", disclaimer=disclaimer
        )
        self._add_key_to_current_property_instance(
            "binding-potential-energy-per-atom",
            binding_potential_energy_per_atom,
            unit="eV",
        )
        self._add_key_to_current_property_instance(
            "binding-potential-energy-per-formula",
            binding_potential_energy_per_formula,
            unit="eV",
        )
        self._add_property_instance_and_common_crystal_genome_keys(
            "mass-density-crystal-npt",
            write_stress=True,
            write_temp=True,
            disclaimer=disclaimer,
        )
        self._add_key_to_current_property_instance(
            "mass-density", self._get_mass_density(), unit="amu/angstrom^3"
        )



from .core import KIMVVTestDriver, override_call_method
from .EquilibriumCrystalStructure.test_driver.test_driver import TestDriver as __EquilibriumCrystalStructure
from .ElasticConstantsCrystal.test_driver.test_driver import TestDriver as __ElasticConstantsCrystal
from .CrystalStructureAndEnergyVsPressure.test_driver.test_driver import TestDriver as __CrystalStructureAndEnergyVsPressure
from .GroundStateCrystalStructure.test_driver.test_driver import TestDriver as __GroundStateCrystalStructure
from .VacancyFormationEnergyRelaxationVolumeCrystal.test_driver.test_driver import TestDriver as __VacancyFormationEnergyRelaxationVolumeCrystal


@override_call_method
class EquilibriumCrystalStructure(__EquilibriumCrystalStructure, KIMVVTestDriver):
    pass


@override_call_method
class ElasticConstantsCrystal(__ElasticConstantsCrystal, KIMVVTestDriver):
    pass


@override_call_method
class CrystalStructureAndEnergyVsPressure(__CrystalStructureAndEnergyVsPressure, KIMVVTestDriver):
    pass


@override_call_method
class GroundStateCrystalStructure(__GroundStateCrystalStructure, KIMVVTestDriver):
    pass


@override_call_method
class VacancyFormationEnergyRelaxationVolumeCrystal(__VacancyFormationEnergyRelaxationVolumeCrystal, KIMVVTestDriver):
    pass


__all__ = [
    "EquilibriumCrystalStructure",
    "ElasticConstantsCrystal",
    "CrystalStructureAndEnergyVsPressure",
    "GroundStateCrystalStructure",
    "VacancyFormationEnergyRelaxationVolumeCrystal",
]



from kim_tools import SingleCrystalTestDriver
from kim_tools.aflow_util.core import get_atom_indices_for_each_wyckoff_orb
import numpy as np
from ase.geometry.cell import cellpar_to_cell
from ase.optimize import FIRE
from scipy.optimize import fmin
import sys
import math
from collections import OrderedDict

KEY_SOURCE_VALUE = 'source-value'
KEY_SOURCE_UNIT = 'source-unit'
KEY_SOURCE_UNCERT = 'source-std-uncert-value'

def V(value, unit = '', uncert = ''):
    # Generate OrderedDict for JSON Dump
    res = OrderedDict([
        (KEY_SOURCE_VALUE, value),
    ])
    if unit != '':
        res.update(OrderedDict([
            (KEY_SOURCE_UNIT, unit),
        ]))
    if uncert != '':
        res.update(OrderedDict([
            (KEY_SOURCE_UNCERT, uncert)
        ]))
    return res

# Parameters for Production
FIRE_MAX_STEPS = 1000
FIRE_UNCERT_STEPS = 20
FIRE_TOL = 1e-3 # absolute
FMIN_FTOL = 1e-6 # relative
FMIN_XTOL = 1e-10 # relative
VFE_TOL = 1e-5 # absolute
MAX_LOOPS = 20
CELL_SIZE_MIN = 3
CELL_SIZE_MAX = 5
COLLAPSE_CRITERIA_VOLUME = 0.1
COLLAPSE_CRITERIA_ENERGY = 0.1
DYNAMIC_CELL_SIZE = True # Increase Cell Size According to lattice structure
EPS = 1e-3

# Extrapolation Parameters
FITS_CNT = [2, 3, 3, 3, 3] # Number of data points used for each fitting
FITS_ORDERS = [
    [0, 3],
    [0, 3],
    [0, 3, 4],
    [0, 3, 5],
    [0, 3, 6],
] # Number of orders included in each fitting
# Fit Results Used (Corresponding to the above)
FITS_VFE_VALUE = 0 # Vacancy Formation Energy
FITS_VFE_UNCERT = [1, 2]
FITS_VRV_VALUE = 0 # Vacancy Relaxation Volume
FITS_VRV_UNCERT = [1, 2]

# Strings for Output
UNIT_ENERGY = 'eV'
UNIT_LENGTH = 'angstrom'
UNIT_ANGLE = 'degree'
UNIT_PRESSURE = 'GPa'
UNIT_VOLUME = UNIT_LENGTH + '^3'

# Disclaimer Uncertainty
DISCLAIMER_THRESHOLD_PERCENT = 1.
DISCLAIMER = f"value has an estimated uncertainty greater than {DISCLAIMER_THRESHOLD_PERCENT}%."

class TestDriver(SingleCrystalTestDriver):
    def _calculate(self, reservoir_info=None, **kwargs):
        self.atoms = self._get_atoms()
        if reservoir_info is None:
            ele = set(self.atoms.get_chemical_symbols())
            reservoir_info = {}
            for e in ele:
                reservoir_info[e] = [{"binding-potential-energy-per-atom":{"source-value": 0}, "prototype-label": {"source-value": 'Not provided'}}]
        self.reservoir_info = reservoir_info 
        prototype_label  = self._SingleCrystalTestDriver__nominal_crystal_structure_npt['prototype-label']['source-value']
        self.prototype_label = prototype_label
        self.equivalent_atoms = get_atom_indices_for_each_wyckoff_orb(prototype_label)
        # store intermediate results for effective vacancy
        self.effective_results = []
        self.bulk_energies = {}

        if DYNAMIC_CELL_SIZE == True:
            numAtoms = self.atoms.get_number_of_atoms()
            factor = math.pow(8 / numAtoms, 0.333)
            self.cell_size_min = int(math.ceil(factor * CELL_SIZE_MIN))
            self.cell_size_max = self.cell_size_min + 2
            print('CELL_SIZE_MIN:', self.cell_size_min)
            print('CELL_SIZE_MAX:', self.cell_size_max)
            print('Smallest System Size:', numAtoms * self.cell_size_min**3)
            print('Largest System Size:', numAtoms * self.cell_size_max**3)

        results = []
        for wkof in self.equivalent_atoms:
            idx = wkof['indices'][0] 
            results.append(self.getResults(idx, len(wkof['indices'])))
        organized_props, uncertainty = self.organize_properties(results)
        for k,v in organized_props.items():
            disclaimer = None
            for u in uncertainty[k]["percent"]:
                if u > DISCLAIMER_THRESHOLD_PERCENT: 
                    disclaimer = f"{uncertainty[k]['prop']} " + DISCLAIMER
                    print ("DISCLAIMER:" + disclaimer)
            self._add_property_instance_and_common_crystal_genome_keys(k,
                                                                   write_stress=False, write_temp=False, disclaimer=disclaimer)
            for k2,v2 in v.items():
                if 'source-unit' in v2:
                    if k2 == uncertainty[k]["prop"]:
                        self._add_key_to_current_property_instance(k2, v2['source-value'], v2['source-unit'], {KEY_SOURCE_UNCERT: uncertainty[k]['values']})
                    else:
                        self._add_key_to_current_property_instance(k2, v2['source-value'], v2['source-unit'])
                else:
                    self._add_key_to_current_property_instance(k2, v2['source-value'])
        
        # combine effective results, extrapolate and write property
        bulkEnergies = [v for k,v in self.bulk_energies.items()]
        sizes =  [k for k,v in self.bulk_energies.items()]
        effectiveUnrelaxedList = []
        effectiveRelaxedList = []
        for idx,be in enumerate(bulkEnergies): 
            effectiveUnrelaxed = 0
            effectiveRelaxed = 0
            for effRes in self.effective_results:
                effectiveUnrelaxed += (effRes['multiplicity']/len(self.atoms) * effRes['unrelaxed'][idx])# - (len(self.atoms) * sizes[idx]**3 - 1) / (len(self.atoms) * sizes[idx]**3) * be
                effectiveRelaxed += (effRes['multiplicity']/len(self.atoms) * effRes['relaxed'][idx])# - (len(self.atoms) * sizes[idx]**3 - 1) / (len(self.atoms) * sizes[idx]**3) * be
            effectiveUnrelaxed -= (len(self.atoms) * sizes[idx]**3 - 1) / (len(self.atoms) * sizes[idx]**3) * be
            effectiveRelaxed -= (len(self.atoms) * sizes[idx]**3 - 1) / (len(self.atoms) * sizes[idx]**3) * be
            effectiveUnrelaxedList.append(effectiveUnrelaxed)
            effectiveRelaxedList.append(effectiveRelaxed)
        uev, rev, ueu, reu  = self.extrapolate(effectiveUnrelaxedList, effectiveRelaxedList)
        disclaimer = None
        if ueu / uev *100 > DISCLAIMER_THRESHOLD_PERCENT:
            disclaimer = "unrelaxed-effective-formation-potential-energy " + DISCLAIMER
            print ("DISCLAIMER: "+ disclaimer)
        self._add_property_instance_and_common_crystal_genome_keys(
            "effective-vacancy-unrelaxed-formation-potential-energy-crystal",
            write_stress=False, 
            write_temp=False,
            disclaimer=disclaimer
        )
        self._add_key_to_current_property_instance("unrelaxed-effective-formation-potential-energy", uev, UNIT_ENERGY, {KEY_SOURCE_UNCERT: ueu})

        disclaimer = None
        if reu / rev *100 > DISCLAIMER_THRESHOLD_PERCENT:
            disclaimer = "relaxed-effective-formation-potential-energy " + DISCLAIMER
            print ("DISCLAIMER: "+ disclaimer)
        self._add_property_instance_and_common_crystal_genome_keys(
            "effective-vacancy-relaxed-formation-potential-energy-crystal", 
            write_stress=False, 
            write_temp=False,
            disclaimer=disclaimer
        )
        self._add_key_to_current_property_instance("relaxed-effective-formation-potential-energy", rev, UNIT_ENERGY, {KEY_SOURCE_UNCERT: reu})
        
        
    def _createSupercell(self, size):
        atoms = self.atoms.copy()
        atoms.set_calculator(self._calc)
        atoms *= (size, size, size)
        return atoms

    def _cellVector2Cell(self, cellVector):
        cell = cellpar_to_cell(cellVector)
        return cell
    
    def _getVFE(self, cellVector, atoms):
        newCell = self._cellVector2Cell(cellVector)
        atoms.set_cell(newCell, scale_atoms = True)
        enAtomsWithVacancy = atoms.get_potential_energy()
        return enAtomsWithVacancy
        
    def _getResultsForSize(self, size, idx):
        # Setup Environment
        unrelaxedCell = self.atoms.get_cell() * size
        atoms = self._createSupercell(size)
        unrelaxedCellVector = atoms.get_cell_lengths_and_angles() 
        numAtoms = atoms.get_number_of_atoms()
        enAtoms = atoms.get_potential_energy()
        self.bulk_energies[size] = enAtoms
        unrelaxedCellEnergy = enAtoms
        unrelaxedCellVolume = np.abs(np.linalg.det(unrelaxedCell))
        print('\nSupercell Size:\n', size)
        print('Unrelaxed Cell:\n', unrelaxedCell)
        print('Unrelaxed Cell Vector:\n', unrelaxedCellVector)
        print('Unrelaxed Cell Energy:\n', unrelaxedCellEnergy)

        # Create Vacancy 
        del atoms[idx]
        enAtomsWithVacancy = atoms.get_potential_energy()

        print('Energy of Unrelaxed Cell With Vacancy:\n', enAtomsWithVacancy)
        enVacancyUnrelaxed = enAtomsWithVacancy - enAtoms + self.chemical_potential

        # Self Consistent Relaxation
        #enVacancy = 0
        effectiveRelaxed = 0

        relaxedCellVector = unrelaxedCellVector
        loop = 0
        while 1:
            # Position Relaxation
            print('==========')
            print('Loop:', loop)
            print('Position Relaxation...')
            dyn = FIRE(atoms)
            dyn.run(fmax = FIRE_TOL, steps = FIRE_MAX_STEPS)
            numSteps = dyn.get_number_of_steps()
            if numSteps >= FIRE_MAX_STEPS:
                print('WARNING: Max number of steps exceeded. Structure may be unstable.')
                # sys.exit(0)
            print('Relaxation Completed. Steps:', numSteps)

            # Cell Size Relaxation
            print('Cell Size Relaxation...')
            tmpCellVector, tmpEnVacancy = fmin(
                self._getVFE,
                relaxedCellVector,
                args = (atoms,),
                ftol = FMIN_FTOL,
                xtol = FMIN_XTOL,
                full_output = True,
            )[:2]

            # Convergence Requirement Satisfied
            if abs(tmpEnVacancy - effectiveRelaxed) < VFE_TOL and dyn.get_number_of_steps() < 1:
                dyn.run(fmax = FIRE_TOL * EPS, steps = FIRE_UNCERT_STEPS)
                tmpCellVector, tmpEnVacancy = fmin(
                    self._getVFE,
                    relaxedCellVector,
                    args = (atoms,),
                    ftol = FMIN_FTOL * EPS,
                    xtol = FMIN_XTOL * EPS,
                    full_output = True,
                )[:2]
                self.VFEUncert = np.abs(tmpEnVacancy - effectiveRelaxed)
                enVacancy = tmpEnVacancy - enAtoms + self.chemical_potential
                oldVolume = np.linalg.det(self._cellVector2Cell(relaxedCellVector))
                newVolume = np.linalg.det(self._cellVector2Cell(tmpCellVector.tolist()))
                self.VRVUncert = np.abs(newVolume - oldVolume)
                relaxedCellVector = tmpCellVector.tolist()
                break

            # Evf = Ev - E0 + mu, where mu is chemical potential of removed element
            enVacancy = tmpEnVacancy - enAtoms + self.chemical_potential
            print ('enVacancy', enVacancy)
            relaxedCellVector = tmpCellVector.tolist()
            # for effective
            effectiveRelaxed = tmpEnVacancy 

            # Check Loop Limit
            loop += 1
            if loop > MAX_LOOPS:
                print('Loops Limit Exceeded. Structure Unstable.')
                sys.exit(0)

            # Output Temporary Result
            relaxedCell = self._cellVector2Cell(relaxedCellVector)
            relaxedCellVolume = np.abs(np.linalg.det(relaxedCell))
            relaxationVolume = unrelaxedCellVolume - relaxedCellVolume
            print('Current VFE:', enVacancy)
            print('Energy of Supercell:', enAtoms)
            print('Unrelaxed Cell Volume:', unrelaxedCellVolume)
            print('Current Relaxed Cell Volume:', relaxedCellVolume)
            print('Current Relaxation Volume:', relaxationVolume)
            print('Current Cell:\n', np.array(self._cellVector2Cell(relaxedCellVector)))

            # Determine Collapse
            if np.abs(relaxationVolume) > COLLAPSE_CRITERIA_VOLUME * unrelaxedCellVolume:
                print('System Collapsed. Volume significantly changed.')
                sys.exit(0)
            if np.abs(enVacancy) > COLLAPSE_CRITERIA_ENERGY * np.abs(enAtoms):
                print('System Collapsed. System Energy significantly changed.')
                sys.exit(0)


        # Print Summary
        print('---------------')
        print('Calculation Completed.')
        print('Number Of Atoms in Supercell:', numAtoms)
        print('Vacancy Formation Energy (relaxed):', enVacancy)
        print('Vacancy Formation Energy (unrelaxed):', enVacancyUnrelaxed)
        print('Unrelaxed Cell Volume:', unrelaxedCellVolume)
        print('Relaxed Cell Volume:', relaxedCellVolume)
        print('Relaxation Volume:', relaxationVolume)
        print('Relaxed Cell Vector:\n', relaxedCellVector)
        print('Unrelaxed Cell Vector:\n', unrelaxedCellVector)
        print('Relaxed Cell:\n', np.array(self._cellVector2Cell(relaxedCellVector)))
        print('Unrelaxed Cell:\n', np.array(self._cellVector2Cell(unrelaxedCellVector)))

        return enVacancyUnrelaxed, relaxedCellVector, enVacancy, relaxationVolume, enAtomsWithVacancy, effectiveRelaxed

    def _getFit(self, xdata, ydata, orders):
        # Polynomial Fitting with Specific Orders
        A = []
        print('\nFit with Size:', xdata)
        print('Orders:', orders)
        for order in orders:
            A.append(np.power(xdata * 1.0, -order))
        A = np.vstack(A).T
        print('Matrix A (Ax = y):\n', A)
        print('Data for Fitting:', ydata)
        res = np.linalg.lstsq(A, ydata, rcond=None)
        print('Fitting Results:', res)
        return res[0]

    def _getValueUncert(self, valueFitId, uncertFitIds, systematicUncert, maxSizeId, dataSource):
        # Get sourceValue and sourceUncert use only certain size and fits
        # Get source value
        valueFitCnt = FITS_CNT[valueFitId]
        sourceValue = dataSource[valueFitId][maxSizeId - valueFitCnt + 1]

        # Get source uncertainty (statistical)
        sourceUncert = 0
        for uncertFitId in uncertFitIds:
            uncertFitCnt = FITS_CNT[uncertFitId]
            uncertValue = dataSource[uncertFitId][maxSizeId - uncertFitCnt + 1]
            sourceUncert = max([abs(uncertValue - sourceValue), sourceUncert])

        # Include systematic error, assuming independent of statistical errors
        sourceUncert = math.sqrt(sourceUncert**2 + systematicUncert**2)
        return sourceValue, sourceUncert

    def getResults(self, idx, mult):
        # grab chemical potential
        # add back isolated atom energy
        if len(set(self.atoms.get_chemical_symbols())) == 1:
            print ("Single element crystal detected-using self as reservoir.")
            self.chemical_potential = self.atoms.get_potential_energy()/len(self.atoms) 
            self.single_element = True
        else:
            self.chemical_potential = self.reservoir_info[self.atoms[idx].symbol][0]["binding-potential-energy-per-atom"]["source-value"] + self.get_isolated_energy_per_atom(self.atoms[idx].symbol) 
            self.single_element = False
        print ('Chemical Potential', self.chemical_potential)


        unitBulk = self.atoms

        # Calculate VFE and VRV for Each Size
        sizes = []
        unrelaxedformationEnergyBySize = []
        formationEnergyBySize = []
        relaxationVolumeBySize = []
        unrelaxedEffectiveBySize = []
        relaxedEffectiveBySize = []
        print('\n[Calculation]')
        for size in range(self.cell_size_min, self.cell_size_max + 1):
            unrelaxedFormationEnergy, relaxedCellVector, relaxedFormationEnergy, relaxationVolume, unrelaxedEffective, relaxedEffective = self._getResultsForSize(size, idx)
            sizes.append(size)
            unrelaxedformationEnergyBySize.append(unrelaxedFormationEnergy)
            formationEnergyBySize.append(relaxedFormationEnergy)
            relaxationVolumeBySize.append(relaxationVolume)
            unrelaxedEffectiveBySize.append(unrelaxedEffective)
            relaxedEffectiveBySize.append(relaxedEffective)

        print('\n[Calculation Results Summary]')
        print('Sizes:', sizes)
        print('Unrelaxed Formation Energy By Size:\n', unrelaxedformationEnergyBySize)
        print('Formation Energy By Size:\n', formationEnergyBySize)
        print('Relaxation Volume By Size:\n', relaxationVolumeBySize)
        self.effective_results.append({'multiplicity': mult, 'unrelaxed': unrelaxedEffectiveBySize, "relaxed": relaxedEffectiveBySize})
        self.sizes = sizes

        # Extrapolate for VFE and VRV of Infinite Size
        print('\n[Extrapolation]')
        naSizes = np.array(sizes)
        naUnrelaxedFormationEnergyBySize = np.array(unrelaxedformationEnergyBySize)
        naFormationEnergyBySize = np.array(formationEnergyBySize)
        naRelaxationVolumeBySize = np.array(relaxationVolumeBySize)
        unrelaxedformationEnergyFitsBySize = []
        formationEnergyFitsBySize = []
        relaxationVolumeFitsBySize = []
        for i in range(0, len(FITS_CNT)):
            cnt = FITS_CNT[i] # Num of Data Points Used
            orders = FITS_ORDERS[i] # Orders Included
            print('Fitting with', cnt, 'points, including orders', orders)
            unrelaxedformationEnergyFits = []
            formationEnergyFits = []
            relaxationVolumeFits = []
            for j in range(0, len(sizes) - cnt + 1):
                print('Fit with data beginning', j)
                xdata = naSizes[j:(j + cnt)]
                unrelaxedformationEnergyFits.append(self._getFit(
                    xdata,
                    naUnrelaxedFormationEnergyBySize[j:(j + cnt)],
                    orders
                )[0])
                formationEnergyFits.append(self._getFit(
                    xdata,
                    naFormationEnergyBySize[j:(j + cnt)],
                    orders
                )[0])
                relaxationVolumeFits.append(self._getFit(
                    xdata,
                    naRelaxationVolumeBySize[j:(j + cnt)],
                    orders
                )[0])
            unrelaxedformationEnergyFitsBySize.append(unrelaxedformationEnergyFits)
            formationEnergyFitsBySize.append(formationEnergyFits)
            relaxationVolumeFitsBySize.append(relaxationVolumeFits)

        # Output Fitting Results
        print('\n[Fitting Results Summary]')
        print('Sizes:', sizes)
        print('Data Points Used:', FITS_CNT)
        print('Orders Included:\n', FITS_ORDERS)
        print('Unrelaxed Formation Energy Fits By Size:\n', unrelaxedformationEnergyFitsBySize)
        print('Formation Energy Fits By Size:\n', formationEnergyFitsBySize)
        print('Relaxation Volume Fits By Size:\n', relaxationVolumeFitsBySize)

        # Obtain Extrapolated Value and Uncertainty
        unrelaxedformationEnergy, unrelaxedformationEnergyUncert = self._getValueUncert(
            FITS_VFE_VALUE,
            FITS_VFE_UNCERT,
            # FMIN_FTOL * formationEnergyBySize[-1],
            self.VFEUncert,
            2,
            unrelaxedformationEnergyFitsBySize,
        )
        formationEnergy, formationEnergyUncert = self._getValueUncert(
            FITS_VFE_VALUE,
            FITS_VFE_UNCERT,
            # FMIN_FTOL * formationEnergyBySize[-1],
            self.VFEUncert,
            2,
            formationEnergyFitsBySize,
        )
        relaxationVolume, relaxationVolumeUncert = self._getValueUncert(
            FITS_VRV_VALUE,
            FITS_VRV_UNCERT,
            # FMIN_XTOL * (self.latticeConsts[0] * CELL_SIZE_MAX)**3,
            self.VRVUncert,
            2,
            relaxationVolumeFitsBySize,
        )

        # Construct Results Dictionary
        unrelaxedformationEnergyResult = OrderedDict([
            ('unrelaxed-formation-potential-energy', V(unrelaxedformationEnergy, UNIT_ENERGY, unrelaxedformationEnergyUncert)),
        ])
        formationEnergyResult = OrderedDict([
            ('relaxed-formation-potential-energy', V(formationEnergy, UNIT_ENERGY, formationEnergyUncert)),
        ])
        relaxationVolumeResult = OrderedDict([
            ('relaxation-volume', V(relaxationVolume, UNIT_VOLUME, relaxationVolumeUncert)),
        ])

        results = {"monovacancy-unrelaxed-formation-potential-energy-crystal": unrelaxedformationEnergyResult, 
                   "monovacancy-relaxed-formation-potential-energy-crystal": formationEnergyResult, 
                   "monovacancy-relaxation-volume-crystal": relaxationVolumeResult}
        return results

    def extrapolate(self, unrelaxed, relaxed):
        naSizes = np.array(self.sizes)
        naUnrelaxed = np.array(unrelaxed)
        naRelaxed = np.array(relaxed)
        unrelaxedEffectiveFormationEnergyFitsBySize = []
        effectiveFormationEnergyFitsBySize = []
        for i in range(0, len(FITS_CNT)):
            cnt = FITS_CNT[i] # Num of Data Points Used
            orders = FITS_ORDERS[i] # Orders Included
            print('Fitting with', cnt, 'points, including orders', orders)
            unrelaxedEffectiveFormationEnergyFits = []
            relaxedEffectiveFormationEnergyFits = []
            for j in range(0, len(self.sizes) - cnt + 1):
                print('Fit with data beginning', j)
                xdata = naSizes[j:(j + cnt)]
                unrelaxedEffectiveFormationEnergyFits.append(self._getFit(
                    xdata,
                    naUnrelaxed[j:(j + cnt)],
                    orders
                )[0])
                relaxedEffectiveFormationEnergyFits.append(self._getFit(
                    xdata,
                    naRelaxed[j:(j + cnt)],
                    orders
                )[0])
            unrelaxedEffectiveFormationEnergyFitsBySize.append(unrelaxedEffectiveFormationEnergyFits)
            effectiveFormationEnergyFitsBySize.append(relaxedEffectiveFormationEnergyFits)

        # Output Fitting Results
        print('\n[Fitting Results Summary]')
        print('Sizes:', self.sizes)
        print('Data Points Used:', FITS_CNT)
        print('Orders Included:\n', FITS_ORDERS)
        print('Unrelaxed Effective Formation Energy Fits By Size:\n', unrelaxedEffectiveFormationEnergyFitsBySize)
        print('Effective Formation Energy Fits By Size:\n', effectiveFormationEnergyFitsBySize)

        # Obtain Extrapolated Value and Uncertainty
        unrelaxedEffectiveFormationEnergy, unrelaxedEffectiveFormationEnergyUncert = self._getValueUncert(
            FITS_VFE_VALUE,
            FITS_VFE_UNCERT,
            # FMIN_FTOL * formationEnergyBySize[-1],
            self.VFEUncert,
            2,
            unrelaxedEffectiveFormationEnergyFitsBySize,
        )
        relaxedEffectiveFormationEnergy, relaxedEffectiveFormationEnergyUncert = self._getValueUncert(
            FITS_VFE_VALUE,
            FITS_VFE_UNCERT,
            # FMIN_FTOL * formationEnergyBySize[-1],
            self.VFEUncert,
            2,
            effectiveFormationEnergyFitsBySize,
        )
        return unrelaxedEffectiveFormationEnergy, relaxedEffectiveFormationEnergy, unrelaxedEffectiveFormationEnergyUncert, relaxedEffectiveFormationEnergyUncert
    def organize_properties(self, results):
        uncertainty = {}
        organized_props = {}
        for r in results:
            for k,v in r.items():
                if k not in organized_props:
                    organized_props[k] = {}
                if k not in uncertainty:
                    uncertainty[k] = {"values": [], "percent": []}
                for k2,v2 in v.items():
                    if k2 not in organized_props[k]:
                        organized_props[k][k2] = {}
                        organized_props[k][k2]['source-value'] = [v2['source-value']]
                    else:
                        organized_props[k][k2]['source-value'].append(v2['source-value'])
                uncertainty[k]['values'].append(v2[KEY_SOURCE_UNCERT])
                uncertainty[k]['percent'].append(v2[KEY_SOURCE_UNCERT] / v2['source-value'] * 100.)
                uncertainty[k]['prop'] = k2
                organized_props[k][k2]['source-unit'] = v2['source-unit'] # must all be same
       
        # get reservoir and host info
        res_info = {}
        host_info = {}
        for idx,i in enumerate(self.equivalent_atoms):
            ele = self.atoms.get_chemical_symbols()[i['indices'][0]]
            if self.single_element:
                res_info[ele] = {
                    'chemical_potential': self.chemical_potential,
                    'prototype_label': self.prototype_label
                }
            else:
                res_info[ele] = {
                    'chemical_potential': self.reservoir_info[ele][0]["binding-potential-energy-per-atom"]["source-value"],
                    'prototype_label': self.reservoir_info[ele][0]["prototype-label"]["source-value"]
                }
            host_info[idx] = {
                'species': ele,
                'coord': self.atoms.get_scaled_positions()[i['indices'][0]], 
                'letter': i['letter']
            }
        for k,v in organized_props.items():
            if k != 'monovacancy-relaxation-volume-crystal': # add reservoir info
                organized_props[k].setdefault('reservoir-chemical-potential', {})['source-value'] = [v['chemical_potential'] for k,v in res_info.items()]
                organized_props[k].setdefault('reservoir-chemical-potential', {})['source-unit'] = UNIT_ENERGY
                organized_props[k].setdefault('reservoir-prototype-label', {})['source-value'] =   [v['prototype_label'] for k,v in res_info.items()]
                # set host info
            organized_props[k].setdefault('vacancy-wyckoff-coordinates', {})['source-value'] = [v['coord'] for k,v in host_info.items()]
            organized_props[k].setdefault('vacancy-wyckoff-species', {})['source-value'] = [v['species'] for k,v in host_info.items()]
            organized_props[k].setdefault('vacancy-wyckoff-letter', {})['source-value'] = [v['letter'] for k,v in host_info.items()]
            organized_props[k].setdefault('host-primitive-cell', {})['source-value'] = self.atoms.get_cell()[:,:]
            organized_props[k].setdefault('host-primitive-cell', {})['source-unit'] = UNIT_LENGTH

        return organized_props, uncertainty

    def _resolve_dependencies(self, material, **kwargs):
        import kimvv
        print("Resolving dependencies...")
        # relax structure
        ecs_test = kimvv.EquilibriumCrystalStructure(self._calc)
        ecs_results = ecs_test(material)
        for result in ecs_results:
            if result["property-id"].endswith("crystal-structure-npt"):
                material_relaxed = result
                break
        # get reservoir info
        gse_test = kimvv.GroundStateCrystalStructure(self._calc)
        reservoir_info = {}
        for ele in result['stoichiometric-species']['source-value']:
            results = gse_test(ele)
            reservoir_info[ele] = results
        kwargs['reservoir_info'] = reservoir_info
        return material_relaxed, kwargs    

if __name__ == "__main__":
    from ase.build import bulk
    from kim_tools import query_crystal_structures
    kim_model_name = "EAM_IMD_SchopfBrommerFrigan_2012_AlMnPd__MO_878712978062_003"
    #kim_model_name = "EAM_Dynamo_AcklandTichyVitek_1987_Ni__MO_977363131043_005"
    list_of_queried_structures = query_crystal_structures(
        kim_model_name=kim_model_name,
        stoichiometric_species=['Al', 'Pd'],
        prototype_label="A2B_cF12_225_c_a",
    )
    print (type(list_of_queried_structures),list_of_queried_structures)
    test = TestDriver(kim_model_name)
    test(list_of_queried_structures[0])
    #test(bulk('Ni'))
    test.write_property_instances_to_file()



"""
Compute elasticity matrix for an arbitrary crystal structure defined by a
periodic cell and basis atom positions.

The following methods for computing the elastic constants are supported:

(I) energy-condensed : Compute elastic constants from the hessian of the
condensed strain energy density, W_eff(eps) = min_d W(eps,d), where d are
the displacements of the internal atoms (aside from preventing rigid-body
translation) and eps is the strain.

(II) stress-condensed : Compute elastic constants from the jacobian of the
condensed stress, sig_eff(eps) = sig(eps,dmin), where dmin = arg min_d
W(eps,d).

(III) energy-full : Compute elastic constants from the hessian of the full
strain energy density, W(eps,d). This involves an algebraic manipulation
to account for the effect of atom relaxation; see eqn (27), in Tadmor et
al, Phys. Rev. B, 59:235-245, 1999.

For well-behaved potentials, all three methods give similar results with
differences beyond the first or second digit due to the numerical differentiation.
The `energy-condensed` and `energy-full` approaches have comparable accuracy,
but full Hessian is *much* slower. The `stress-condensed` approach is
significantly faster than `energy-condensed`, but is less accurate.

For potentials with a rough energy landscape due to sharp cut-offs or
electrostatics, `stress-condensed` is sometimes better behaved.

The default approach is `energy-condensed`.

=== Theory ===

All three approaches involve numerical differentiation of energy or stress
expressions in terms of strains represented in Voigt notation. This is requires
the introduction of factors to account for the Voigt form as explained below.

The code assumes a Voigt notation with the following ordering:
1 = 11,  2 = 22,  3 = 33,  4 = 23,  5 = 13,  6 = 12  ---------------(1)

In this notation, the linear elastic stress-strain relation for infinitesimal
defomation is give by:

sig_m = C_mn eps_n -------------------------------------------------(2)

where

sig_1 = sig_11,  sig_2 = sig_22,  sig_3 = sig_33
sig_4 = sig_23,  sig_5 = sig_13,  sig_6 = sig_12

eps_1 = eps_11,  eps_2 = eps_22,  eps_3 = eps_33
eps_4 = gam_23,  eps_5 = gam_13,  eps_6 = gam_12 -------------------(3)

and gam_ij = 2 eps_ij.

In the above sig_ij and eps_ij are the components of the stress and strain
tensors in a Cartesian basis. The same basis is used to define the orientation
of the periodic cell and basis atom positions.

=== Energy-based calculation of elastic constants ===

The strain energy density of a linear elastic meterial is

W = 1/2 c_ijkl eps_ij eps_kl ---------------------------------------(4)

where c_ijkl are the components of the 4th-order elasticity tensor.
In Voigt notation strain energy density has the form:

W = 1/2 C_mn eps_m eps_n -------------------------------------------(5)

where C_mn are the components of elasticity matrix. The symmetries in the
summation of Eqn (4) (which includes 81 terms) are captured by the factors
of two in the 4, 5, 6 components of eps_* in Eqn (3).

Eqn (4) implies that the elasticity tensor components are given by

c_ijkl = d^2 W / d eps_ij d eps_kl ---------------------------------(6)

Substituting into Eqn (6), the definition of W in Eqn (5), a chain rule
must be applied:

c_ijkl = (d^2 W / d eps_m d eps_n )(d eps_m / d eps_ij)(d eps_n / d eps_kl)

                                                                    (7)

Transforming to Voigt notation and accounting for the factors of two
in the shear terms in Eqn (3), we have

C_mn = d^2 W / d eps_m d eps_n               for m,n <=3

C_mn = (1/2) d^2 W / d eps_m d eps_n         for m<=3, n>3  or
                                                 m>3, n<=3

C_mn = (1/4) d^2 W / d eps_m d eps_n         for m,n > 3

=== Stress-based calculation of elastic constants ===

An alternative aproach to computing the elastic constants is to use the
stress-strain relation in Eqn (2), which in full tensorial notation is

sig_ij = c_ijkl eps_kl ---------------------------------------------(8)

This relation implies that

c_ijkl = d sig_ij / d eps_kl ---------------------------------------(9)

Substituting in Eqn (2) for the stress and applying the chain rule,

c_ijkl = (d sig_m / d eps_n)(d eps_n / d eps_kl) ------------------(10)

where m is the Voigt index for ij. Transforming to Voigt notation and
accounting for the factors of two in the shear terms in Eqn (3), we have

C_mn = d sig_m / d eps_n                     for m,n <=3

C_mn = (1/2) d sig_m / d eps_n               for n > 3

Revisions:

2019/04/13 Ellad Tadmor added ability to do diamond)
2022/02/01 Chloe Zeller (lammps compatibility - specific usage case)
2022/02/22 Ellad Tadmor generalized to arbitrary crystal structures
2024/05/08 Ilia Nikiforov Symmetry checking and refactoring for robust Crystal Genome
operation
2025/07/19 Ilia Nikiforov Refactor for Elastic Constants Test Driver 001. Move
error checking and iteration out of this file, leaving that to the calling routine.
Allow option of symmetry restriction.

"""

import logging
import numpy as np
import numpy.typing as npt
from ase.atoms import Atoms
import numdifftools as ndt
from numdifftools.step_generators import MaxStepGenerator
import math
from typing import Union, Tuple, Dict
from kim_tools import minimize_wrapper
from kim_tools.symmetry_util.core import (
    fractional_to_cartesian_itc_rotation_from_ase_cell,
    FixProvidedSymmetry,
    get_primitive_genpos_ops,
)
from ase.optimize.optimize import Optimizer
from ase.optimize.lbfgs import LBFGSLineSearch


logger = logging.getLogger(__name__)
logging.basicConfig(filename="kim-tools.log", level=logging.INFO, force=True)


def energy_hessian_add_prefactors(
    hessian: npt.ArrayLike, hessian_error_estimate: npt.ArrayLike
) -> Tuple[npt.ArrayLike, npt.ArrayLike]:
    """
    Hessian was computed as d^2 W / d eps_m d eps_n.
    As noted in the module doc string, a prefactor is required for some terms
    to obtain the elastic constants

    Returns:
        * 6x6 elasticity matrix in Voigt order
        * 6x6 error estimate in Voigt order
    """
    elastic_constants = np.zeros(shape=(6, 6))
    error_estimate = np.zeros(shape=(6, 6))
    for m in range(6):
        if m < 3:
            factm = 1.0
        else:
            factm = 0.5
        for n in range(6):
            if n < 3:
                fact = factm
            else:
                fact = factm * 0.5
            elastic_constants[m, n] = fact * hessian[m, n]
            error_estimate[m, n] = fact * hessian_error_estimate[m, n]
    return elastic_constants, error_estimate


class ElasticConstants(object):
    """
    Compute the elastic constants of an arbitrary crystal
    through numerical differentiation
    """

    def __init__(
        self,
        atoms: Atoms,
        sgnum: int = 1,
        fmax: float = 1e-5,
        steps: int = 200,
        algorithm: Optimizer = LBFGSLineSearch,
        opt_kwargs: Dict = {},
    ):
        """
        Class containing data and routines for elastic constant calculations

        Parameters:
            atoms:
                ASE atoms object with calculator attached. This object
                contains the structure for which the elastic constant will be
                computed.
            sgnum:
                The optimization will be symmetry-restricted based on
                this space group number. In order for this to work correctly,
                the unit cell must be in the orientation defined in
                doi.org/10.1016/j.commatsci.2017.01.017. For each deformation
                step, the symmetry elements broken by the deformation will be
                disabled. Use a value of 1 (default) to disable
                symmetry restriction
            fmax:
                Force convergence tolerance (the magnitude of the force on each
                atom must be less than this for convergence)
            steps:
                Maximum number of iterations for the minimization
            algorithm:
                ASE optimizer algorithm
            opt_kwargs:
                Dictionary of kwargs to pass to optimizer

        """
        self.atoms = atoms
        self.natoms = atoms.get_global_number_of_atoms()
        self.sgnum = sgnum
        # Store the original reference cell structure and volume, and atom
        # positions.
        self.o_cell = self.atoms.get_cell()
        self.o_volume = self.atoms.get_volume()
        self.refpositions = self.atoms.get_positions()
        self.fmax = fmax
        self.steps = steps
        self.algorithm = algorithm
        self.opt_kwargs = opt_kwargs

    def voigt_to_matrix(self, voigt_vec: npt.ArrayLike) -> npt.ArrayLike:
        """
        Convert a voigt notation vector to a matrix

        Parameters:
            voigt_vec:
                A numpy array containing the six strain components in
                Voigt ordering (xx, yy, zz, yz, xz, xy)

        Returns:
            matrix:
               A 3x3 numpy array containg the strain tensor components
        """
        matrix = np.zeros((3, 3))
        matrix[0, 0] = voigt_vec[0]
        matrix[1, 1] = voigt_vec[1]
        matrix[2, 2] = voigt_vec[2]
        matrix[tuple([[1, 2], [2, 1]])] = voigt_vec[3]
        matrix[tuple([[0, 2], [2, 0]])] = voigt_vec[4]
        matrix[tuple([[0, 1], [1, 0]])] = voigt_vec[5]
        return matrix

    def get_energy_from_positions(self, pos: npt.ArrayLike) -> float:
        """
        Compute the potential energy of the configuration, given a set of positions for
        the internal atoms.

        Parameters:
            pos:
                A numpy array containing the positions of all atoms in a
                flat concatenated form

        Returns:
            energy:
               Potential energy of the configuration
        """
        self.atoms.set_positions(np.reshape(pos, (self.natoms, 3)))
        energy = self.atoms.get_potential_energy()

        return energy

    def get_gradient_from_positions(self, pos: npt.ArrayLike) -> npt.ArrayLike:
        """
        Compute the gradient of the configuration's potential energy, given a set of
        positions for the internal atoms.

        Parameters:
            pos:
                A numpy array containing the positions of all atoms in a
                flat concatenated form

        Returns:
            gradient:
               The gradient of the potential energy of the force (the negative
               of the forces) in a flat concatenated form
        """
        self.atoms.set_positions(np.reshape(pos, (self.natoms, 3)))
        forces = self.atoms.get_forces()
        return -forces.flatten()

    def get_energy_from_strain_and_atom_displacements(
        self, strain_and_disps_vec: npt.ArrayLike
    ) -> float:
        """
        Compute reference strain energy density for a given applied strain
        and internal atom positions for all but one atom constrained to
        prevent rigid-body translation.

        Parameters:
            strain_and_disps_vec:
                A numpy array of length 6+(natoms-1)*3 containing the
                strain components in Voigt order and the free internal
                atom degrees of freedom

        Returns:
            energy density:
                Potential energy of the configuration divided by its reference
                (unstrained) volume
        """
        if self.natoms < 2:
            return self.get_energy_from_strain(strain_and_disps_vec)

        # Set atom positions of the last N-1 atoms to reference positions
        # plus displacements, keeping first atom fixed
        defpositions = self.refpositions.copy()
        for i in range(self.natoms - 1):
            disp = strain_and_disps_vec[(6 + i * 3) : (9 + i * 3)]
            defpositions[i + 1] += disp
        self.atoms.set_positions(defpositions)

        # Apply strain to cell scaling the atom positions
        self.atoms.set_cell(self.o_cell, scale_atoms=False)
        strain_vec = strain_and_disps_vec[0:6]
        strain_mat = self.voigt_to_matrix(strain_vec)
        old_cell = self.o_cell
        new_cell = old_cell + np.dot(old_cell, strain_mat)
        self.atoms.set_cell(new_cell, scale_atoms=True)

        # Compute energy
        energy = self.atoms.get_potential_energy()

        return energy / self.o_volume

    def get_elasticity_matrix_and_error_energy_full(
        self, step: Union[MaxStepGenerator, float]
    ) -> Tuple[npt.ArrayLike, npt.ArrayLike]:
        """
        Compute elastic constants from the full Hessian
        relative to both strains and internal atom degrees of
        freedom. This is followed by an algebraic manipulation to
        account for the effect of atom relaxation on elastic
        constants.

        Returns:
            * 6x6 elasticity matrix in Voigt order
            * 6x6 error estimate in Voigt order
        """
        hess = ndt.Hessian(
            self.get_energy_from_strain_and_atom_displacements,
            step=step,
            full_output=True,
        )
        fullhessian, info = hess(np.zeros(6 + (self.natoms - 1) * 3, dtype=float))
        fullhessian_error_estimate = info.error_estimate
        # Separate full Hessian into blocks
        # (eps-eps, eps-disp, disp-disp)
        hess_ee = fullhessian[0:6, 0:6]
        hess_ed = fullhessian[0:6, 6:]
        hess_dd = fullhessian[6:, 6:]
        # Invert disp-disp block
        hess_dd_inv = np.linalg.inv(hess_dd)
        # Compute hessian accounting for basis atom relaxation. based
        # on Eqn (27) in Tadmor et al, Phys. Rev. B, 59:235-245, 1999.
        hessian = hess_ee - np.dot(np.dot(hess_ed, hess_dd_inv), np.transpose(hess_ed))
        # TODO: Figure out how to estimate error of Hessian based on
        #      full Hessian errors in fullhessian_error_estimate
        #      Now just taken errors for hess_ee which is incorrect.
        hessian_error_estimate = info.error_estimate[0:6, 0:6]
        return energy_hessian_add_prefactors(hessian, hessian_error_estimate)

    def get_energy_from_strain(self, strain_vec: npt.ArrayLike) -> float:
        """
        Compute reference strain energy density for a given applied strain.

        Parameters:
            strain_vec:
                A numpy array of length 6 containing the strain components in
                Voigt order

        Returns:
            energy density:
                Potential energy of the configuration divided by its reference
                (unstrained) volume
        """
        logger.info(f"Strain vector: {strain_vec}")

        self.atoms.set_cell(self.o_cell, scale_atoms=False)
        self.atoms.set_positions(self.refpositions)
        strain_mat = self.voigt_to_matrix(strain_vec)

        # Find the symmetry operations for the requested space group
        full_symmetry_ops = get_primitive_genpos_ops(self.sgnum)
        # Knock out things that don't commute with the deformation
        strain_norm = np.linalg.norm(strain_vec)
        # Everything commutes with the identity, no need to change symm at all
        if np.isclose(strain_norm, 0.0):
            symm_ops = full_symmetry_ops
        else:
            symm_ops = []
            deform_direction_mat = np.eye(3) + strain_mat / strain_norm
            for op in full_symmetry_ops:
                cart_rot = fractional_to_cartesian_itc_rotation_from_ase_cell(
                    op["W"], self.o_cell
                )
                if np.allclose(
                    cart_rot @ deform_direction_mat,
                    deform_direction_mat @ cart_rot,
                ):
                    symm_ops.append(op)
        symm = FixProvidedSymmetry(self.atoms, symm_ops)

        old_cell = self.o_cell
        new_cell = old_cell + np.dot(old_cell, strain_mat)
        self.atoms.set_cell(new_cell, scale_atoms=True)

        if self.natoms > 1:
            minimize_wrapper(
                self.atoms,
                fmax=self.fmax,
                steps=self.steps,
                variable_cell=False,
                logfile=None,
                algorithm=self.algorithm,
                fix_symmetry=symm,
                opt_kwargs=self.opt_kwargs,
            )
            energy = self.atoms.get_potential_energy()
        else:
            energy = self.atoms.get_potential_energy()

        # check that symmetry restriction is working properly and not
        # preventing cell from changing
        assert np.allclose(self.atoms.get_cell(), new_cell)

        return energy / self.o_volume

    def get_elasticity_matrix_and_error_energy_condensed(
        self, step: Union[MaxStepGenerator, float]
    ) -> Tuple[npt.ArrayLike, npt.ArrayLike]:
        """
        Compute elastic constants from the Hessian
        of the condensed strain energy density (i.e. the enregy for
        a given strain is relaxed with respect to internal atom
        positions)

        Returns:
            * 6x6 elasticity matrix in Voigt order
            * 6x6 error estimate in Voigt order
        """
        hess = ndt.Hessian(self.get_energy_from_strain, step=step, full_output=True)
        hessian, info = hess(np.zeros(6, dtype=float))
        return energy_hessian_add_prefactors(hessian, info.error_estimate)

    def get_stress_from_strain(self, strain_vec: npt.ArrayLike) -> npt.ArrayLike:
        """
        Compute stress for a given applied strain.

        Parameters:
            strain_vec:
                A numpy array of length 6 containing the strain components in
                Voigt order

        Returns:
            stress:
                Cauchy stress of the configuration
        """
        # Call get_energy_from_strain to strain the
        # atoms
        _ = self.get_energy_from_strain(strain_vec)
        stress = self.atoms.get_stress()
        return stress

    def get_elasticity_matrix_and_error_stress(
        self, step: Union[MaxStepGenerator, float]
    ) -> Tuple[npt.ArrayLike, npt.ArrayLike]:
        """
        Compute elastic constants from the Jacobian
        of the condensed stress (i.e. the stress for a given strain
        where the energy is relaxed with respect to internal atom
        positions)

        Returns:
            * 6x6 elasticity matrix in Voigt order
            * 6x6 error estimate in Voigt order
        """
        jac = ndt.Jacobian(self.get_stress_from_strain, step=step, full_output=True)
        hessian, info = jac(np.zeros(6, dtype=float))
        hessian_error_estimate = info.error_estimate

        elastic_constants = np.zeros(shape=(6, 6))
        error_estimate = np.zeros(shape=(6, 6))

        # Hessian was computed as d sig_m / d eps_n.
        # As noted in the module doc string, a prefactor is required for some terms
        # to obtain the elastic constants
        for m in range(6):
            for n in range(6):
                if n < 3:
                    fact = 1.0
                else:
                    fact = 0.5
                elastic_constants[m, n] = fact * hessian[m, n]
                error_estimate[m, n] = fact * hessian_error_estimate[m, n]
        # The elastic constants matrix should be symmetric, however due to
        # numerical precision issues in the stress components, in general,
        # d sig m / d eps_n will not equal d sig_n / d eps_m.
        # To address this, symmetrize the elastic constants matrix.
        for m in range(5):
            for n in range(m + 1, 6):
                con = 0.5 * (elastic_constants[m, n] + elastic_constants[n, m])
                # Variances of a linear combination sum as the square of the
                # coefficient
                err = math.sqrt(
                    0.25 * (error_estimate[m, n] ** 2 + error_estimate[n, m] ** 2)
                )
                elastic_constants[m, n] = con
                elastic_constants[n, m] = con
                error_estimate[m, n] = err
                error_estimate[n, m] = err

        return elastic_constants, error_estimate

    def results(
        self,
        method: str = "energy-condensed",
        step: Union[MaxStepGenerator, float] = MaxStepGenerator(
            base_step=1e-4,
            num_steps=14,
            use_exact_steps=True,
            step_ratio=1.6,
            offset=0,
        ),
    ) -> Tuple[
        npt.ArrayLike,
        npt.ArrayLike,
        str,
    ]:
        """
        Compute the elastic constants of a crystal using numerical differentiation

        Parameters:
            method:
                Select method for computing the elastic constants. The following
                methods are supported:
                'energy-condensed' : Compute elastic constants from the Hessian
                    of the condensed strain energy density (i.e. the enregy for
                    a given strain is relaxed with respect to internal atom
                    positions)
                'stress-condensed' : Compute elastic constants from the Jacobian
                    of the condensed stress (i.e. the stress for a given strain
                    where the energy is relaxed with respect to internal atom
                    positions)
                'energy-full' : Compute elastic constants from the full Hessian
                    relative to both strains and internal atom degrees of
                    freedom. This is followed by an algebraic manipulation to
                    account for the effect of atom relaxation on elastic
                    constants.
                In general, 'energy-condensed' is the preferred method.
                The 'stress-condensed' method is much faster, but generally less
                accurate. The 'energy-full' method has accuracy comparable
                to 'energy-condensed' but tends to be much slower due to the
                larger Hessian matrix that has to be computed.
            step:
                Step(s) to pass to the numdifftools routines
        Returns:
            *   A 6x6 numpy array containing the elastic constants matrix
                in Voigt ordering. This is the full matrix of derivatives
                returned by numdifftools, with appropriate prefactors added
                for Voigt notation, but not yet corrected for material symmetry.
                Units are the default returned by the calculator.
            *   A 6x6 numpy array containing the 95% error in the elastic
                constants returned by numdifftools, with appropriate prefactors added
                for Voigt notation. Same units as elastic_constants.
            *   A summary of the run and any issues
        """
        if method == "stress-condensed":
            get_elasticity_matrix = self.get_elasticity_matrix_and_error_stress
        elif method == "energy-condensed":
            get_elasticity_matrix = (
                self.get_elasticity_matrix_and_error_energy_condensed
            )
        elif method == "energy-full":
            get_elasticity_matrix = self.get_elasticity_matrix_and_error_energy_full
        else:
            raise RuntimeError(
                "Unknown computation method. Supported methods: "
                "'energy-condensed','stress-condensed','energy-full'"
            )

        elastic_constants, elastic_constants_error_estimate = get_elasticity_matrix(
            step
        )

        message = (
            f"\nMethod: {method}\nStep generator: {step}\n\n"
            + "\nRaw elastic constants [ASE units]:\n"
            + np.array_str(
                elastic_constants,
                precision=5,
                max_line_width=100,
                suppress_small=True,
            )
            + "\n\n"
        )

        if isinstance(step, float):
            # When given a single float step,
            # we cannot give a meaningful error estimate
            elastic_constants_error_estimate = np.full((6, 6), np.nan)
        else:
            message = (
                message
                + "\n95%% Error estimate [ASE units]:\n"
                + np.array_str(
                    elastic_constants_error_estimate,
                    precision=5,
                    max_line_width=100,
                    suppress_small=True,
                )
                + "\n\n"
            )

        print()
        print("Summary of completed elastic constants calculation:")
        print(message)

        return (
            elastic_constants,
            elastic_constants_error_estimate,
            message,
        )



from kim_tools import SingleCrystalTestDriver
from ase import Atoms
from typing import Any, Optional, List, Union, Dict, IO
# try importing as ppart of module otherwise script like
try:
    from .reference_map import reference_map
except ImportError:
    from reference_map import reference_map
class TestDriver(SingleCrystalTestDriver):
    # keep same signature so __call__ works
    def _setup(
        self,
        material: str,
        cell_cauchy_stress_eV_angstrom3: Optional[List[float]] = None,
        temperature_K: float = 0,
        **kwargs,
    ) -> None:
        """
        Args:
            material:
                Material in this case is simply a str containing a single element type,
                e.g., 'Au'.
            cell_cauchy_stress_eV_angstrom3:
                Cauchy stress on the cell in eV/angstrom^3 (ASE units) in
                [xx, yy, zz, yz, xz, xy] format. This is a nominal variable, and this
                class simply provides recordkeeping of it. It is up to derived classes
                to implement actually imposing this stress on the system.
            temperature_K:
                The temperature in Kelvin. This is a nominal variable, and this class
                simply provides recordkeeping of it. It is up to derived classes to
                implement actually setting the temperature of the system.
        """
        if cell_cauchy_stress_eV_angstrom3 is None:
            cell_cauchy_stress_eV_angstrom3 = [0, 0, 0, 0, 0, 0]

    def _calculate(self, reference_info, temperature_K=None, cell_cauchy_stress_eV_angstrom3=None, **kwargs):
        '''
        Args:
            reference_info:
                Result from a KIM query or EquilibriumCrystalStructure test
                Result will be in form {'<element symbol>': [binding-potential-energy-crystal instance]}
            cell_cauchy_stress_eV_angstrom3:
                Cauchy stress on the cell in eV/angstrom^3 (ASE units) in
                [xx, yy, zz, yz, xz, xy] format. This is a nominal variable, and this
                class simply provides recordkeeping of it. It is up to derived classes
                to implement actually imposing this stress on the system.
            temperature_K:
                The temperature in Kelvin. This is a nominal variable, and this class
                simply provides recordkeeping of it. It is up to derived classes to
                implement actually setting the temperature of the system.
        '''
        # Set reference structure and property instance
        if temperature_K is None:
            temperature_K = 0
        if cell_cauchy_stress_eV_angstrom3 is None:
            cell_cauchy_stress_eV_angstrom3 = [0, 0, 0, 0, 0, 0]

        result = list(reference_info.values())[0][0]
        self._SingleCrystalTestDriver__nominal_crystal_structure_npt = result
        self._SingleCrystalTestDriver__nominal_crystal_structure_npt["temperature"] = {
            "source-value": temperature_K,
            "source-unit": "K",
        }
        self._SingleCrystalTestDriver__nominal_crystal_structure_npt["cell-cauchy-stress"] = {
            "source-value": cell_cauchy_stress_eV_angstrom3,
            "source-unit": "eV/angstrom^3",
        }
        self._add_property_instance_and_common_crystal_genome_keys(
                "elemental-crystal-ground-state",
                write_stress=False, write_temp=False
                )
        self._add_key_to_current_property_instance(
                "binding-potential-energy-per-atom",
                result["binding-potential-energy-per-atom"]["source-value"],
                result["binding-potential-energy-per-atom"]["source-unit"]
                )

    def _resolve_dependencies(self, material, **kwargs):
        '''
        Take reference element and map to reference structure
        Compute EquilibriumCrystalStructure and format binding-energy-crystal result
        '''
        import kimvv
        print ("Resolving dependencies...")
        energy = 0
        lowest_res = None
        if isinstance(material,Atoms):
            assert (len(set(material.get_chemical_symbols())) == 1)
            material = material.get_chemical_symbols()[0]
        refs = reference_map[material]
        for r in refs:
            ecs_test = kimvv.EquilibriumCrystalStructure(self._calc)
            ecs_test(r)
            res = ecs_test.property_instances[1]
            if res["binding-potential-energy-per-atom"]["source-value"] < energy:
                energy = res["binding-potential-energy-per-atom"]["source-value"]
                lowest_res = res
        kwargs['reference_info'] = {material: [res]}
        return material, kwargs



import json
import warnings
from typing import Union

import pytest
from ase.build import bulk
from ase.calculators.calculator import Calculator
from ase.calculators.lj import LennardJones
from kim_tools import KIMTestDriver

import kimvv

with open("test_inputs.json") as f:
    DRIVERS = json.load(f)

# Test on FCC Au
MODELS = [
    "LJ_ElliottAkerson_2015_Universal__MO_959249795837_003",
    "Sim_LAMMPS_ADP_StarikovGordeevLysogorskiy_2020_SiAuAl__SM_113843830602_000",
    LennardJones(sigma=2.42324, epsilon=2.30580, rc=9.69298),
]

test_tuples = [(driver, model) for driver in DRIVERS for model in MODELS]


@pytest.mark.parametrize("td_name,model", test_tuples)
@pytest.mark.filterwarnings("ignore:(?!WARNING Your )")
def test_test_driver(td_name: str, model: Union[str, Calculator]) -> None:
    """
    Run ``td_name`` with ``model`` on and confirm that it returns at least one
    instance of the properties it claims to return, and no others.

    Args:
        td_name:
            The name of the class of Test Driver to run.
        model:
            The model to use.
    """
    # Start with FCC Au
    atoms = bulk("Au")

    TestDriver = getattr(kimvv, td_name)

    td_kwargs = DRIVERS[td_name]

    td = TestDriver(model)

    try:
        results = td(atoms, **td_kwargs)

        # Should return at least something
        assert len(results) > 0

        # If we have properties in our kimspec, check that the test driver
        # only reports those
        if "properties" in td.get_kimspec():
            properties = td.get_kimspec()["properties"]
            for result in results:
                assert result["property-id"] in properties
        else:
            warnings.warn(
                "WARNING Your kimspec.edn did not contain a 'properties' key. "
                "this is acceptable, but I cannot test thhat the reported "
                "properties are correct."
            )
    except KIMTestDriver.NonKIMModelError:
        warnings.warn(
            "WARNING Your Test Driver is unable to run with non-KIM calculators "
            "because it requests the kim_model_name property. This is acceptable, "
            "but only if strictly necessary (e.g. needing a KIM model name to run a "
            "LAMMPS calculation)"
        )



import os
from math import exp, floor, log
from typing import Dict, List, Optional, Tuple

import numpy as np
from ase.atoms import Atoms
from ase.data import atomic_masses
from ase.filters import FrechetCellFilter, UnitCellFilter
from ase.optimize.lbfgs import LBFGSLineSearch
from ase.optimize.optimize import Optimizer
from ase.units import GPa, kg, m
from kim_tools import AFLOW, SingleCrystalTestDriver
from kim_tools.symmetry_util.core import FixProvidedSymmetry
from matplotlib import pyplot as plt

CONFIG_TPL = "instance_{prop_idx}_{p_gpa}_GPa.poscar"


class StopScanError(Exception):
    pass


class VolumeMockConstraint:
    """
    "Constraint" that just raises an error if volume is exceeded.
    This is better than using a stopping criterion in an irun loop, because
    it allows stopping in the middle of a linesearch
    """

    def __init__(self, min_volume, max_volume):
        self.min_volume = min_volume
        self.max_volume = max_volume

    def adjust_positions(self, atoms, newpositions):
        pass

    def adjust_forces(self, atoms, forces):
        pass

    def adjust_cell(self, atoms, cell):
        if cell.volume > self.max_volume or cell.volume < self.min_volume:
            raise StopScanError(
                "Volume exceeded prescribed range, stopping pressure scan."
            )


class TestDriver(SingleCrystalTestDriver):
    def _minimize_volume_limited(
        self,
        atoms: Atoms,
        scalar_pressure: float,
    ) -> bool:
        """
        Use LBFGSLineSearch (default) to Minimize cell energy with respect to cell
        shape and internal atom positions, with parameters inherited
        from the class instance.

        LBFGSLineSearch convergence behavior is as follows:

        - The solver returns True if it is able to converge within the optimizer
        iteration limits (which can be changed by the ``steps`` argument passed
        to ``run``), otherwise it returns False.
        - The solver raises an exception in situations where the line search cannot
        improve the solution, typically due to an incompatibility between the
        potential's values for energy, forces, and/or stress.

        This routine attempts to minimizes the energy until the force and stress
        reduce below specified tolerances given a provided limit on the number of
        allowed steps. The code returns when convergence is achieved or no
        further progress can be made, either due to reaching the iteration step
        limit, or a stalled minimization due to line search failures.

        Args:
            atoms:
                Atomic configuration to be minimized.
            scalar_pressure:
                The scalar pressure to apply

        Returns:
            Whether the minimization succeeded
        """
        atoms_wrapped = self.cell_filter(
            atoms, scalar_pressure=scalar_pressure, **self.min_flt_kwargs
        )
        opt = self.min_algorithm(atoms_wrapped, logfile="-", **self.min_opt_kwargs)
        try:
            converged = opt.run(fmax=self.min_fmax, steps=self.min_steps)
            iteration_limits_reached = not converged
            minimization_stalled = False
        except RuntimeError as e:
            if str(e) == "LineSearch failed!":
                minimization_stalled = True
                iteration_limits_reached = False
                print(
                    "LineSearch failed."
                    "Assuming it is non-fatal and accepting result with disclaimer."
                )
            else:
                raise e

        print(
            "Minimization "
            + (
                "stalled"
                if minimization_stalled
                else "stopped" if iteration_limits_reached else "converged"
            )
            + " after "
            + (
                ("hitting the maximum of " + str(self.min_steps))
                if iteration_limits_reached
                else str(opt.nsteps)
            )
            + " steps."
        )

        if minimization_stalled or iteration_limits_reached:
            print("Final forces:")
            print(atoms.get_forces())
            print("Final stress:")
            print(atoms.get_stress())
            return False
        else:
            return True

    def _do_scan(self, backward: bool) -> Tuple[
        List[float],
        List[float],
        List[float],
        List[float],
        List[float],
        List[float],
        List[float],
        List[List[float]],
    ]:
        """
        Backward: linear scan from 0 to -self.max_steps
        Forward: exp scan from 0 to max_steps, with the pressure being
        exp(i*self.exp_step) * self.pressure_step_gpa
        """
        binding_potential_energy_per_atom = []
        binding_potential_energy_per_formula = []
        volume_per_atom = []
        volume_per_formula = []
        density_kg_m3 = []
        pressure_gpa = []
        a = []
        parameter_values = []
        if backward:
            iterrange = reversed(range(-self.max_pressure_steps, 1))
        else:
            iterrange = range(self.max_pressure_steps)

        # Get original atoms
        # Could use self._get_atoms() but no reason
        # to call AFLOW every time
        atoms = self.original_atoms.copy()
        atoms.calc = self._calc
        sgnum = self.get_nominal_space_group_number()
        atoms.set_constraint(
            [
                FixProvidedSymmetry(atoms, sgnum),
                VolumeMockConstraint(self.min_volume, self.max_volume),
            ]
        )
        num_atoms = len(atoms)
        num_atoms_in_formula = sum(self.get_nominal_stoichiometry())
        prototype_label = self.get_nominal_prototype_label()
        aflow = AFLOW()
        for i in iterrange:
            if backward:
                current_pressure_gpa = i * self.pressure_step_gpa
            else:
                current_pressure_gpa = exp(i * self.exp_step) * self.pressure_step_gpa
            print(f"\nPressure: {current_pressure_gpa} GPa\n")
            current_pressure = current_pressure_gpa * GPa
            try:
                # Relax with fixed pressure
                if not self._minimize_volume_limited(atoms, current_pressure):
                    if self.disclaimer is None:
                        self.disclaimer = (
                            "Relaxation failed to reach force tolerance of "
                            f"{self.min_fmax} eV/angstrom at "
                            f"{current_pressure_gpa} GPa"
                        )
                    else:
                        self.disclaimer += f", {current_pressure_gpa} GPa"
                # redetect current crystal structure first, as if this fails we do not
                # report anything. We need this to write the file in the correct
                # orientation later
                curr_param_values = aflow.solve_for_params_of_known_prototype(
                    atoms=atoms,
                    prototype_label=prototype_label,
                    match_library_proto=False,
                )
                # get other quantities
                potential_energy = atoms.get_potential_energy()
                volume = atoms.get_volume()

                # write config
                # property instances are 1-based, but we have not initialized the
                # newest one yet
                prop_idx = len(self.property_instances) + 1
                filename = os.path.join(
                    "output",
                    CONFIG_TPL.format(p_gpa=current_pressure_gpa, prop_idx=prop_idx),
                )

                # Regenerate poscar from parameter values instead of
                # just self.write_atoms -- otherwise they won't be in
                # standard orientation
                aflow.write_poscar_from_prototype(
                    prototype_label=prototype_label,
                    species=self.get_nominal_stoichiometric_species(),
                    parameter_values=curr_param_values,
                    output_file=filename,
                )

                # report things
                current_volume_per_atom = volume / num_atoms
                current_binding_potential_energy_per_atom = potential_energy / num_atoms
                volume_per_atom.append(current_volume_per_atom)
                volume_per_formula.append(
                    current_volume_per_atom * num_atoms_in_formula
                )
                # compute density
                mass = 0.0
                for atomic_number in atoms.get_atomic_numbers():
                    mass += atomic_masses[atomic_number]
                mass_kg = mass / kg
                volume_m3 = volume / m**3
                density_kg_m3.append(mass_kg / volume_m3)
                binding_potential_energy_per_atom.append(
                    current_binding_potential_energy_per_atom
                )
                binding_potential_energy_per_formula.append(
                    current_binding_potential_energy_per_atom * num_atoms_in_formula
                )
                a.append(curr_param_values[0])
                parameter_values.append(curr_param_values[1:])
                pressure_gpa.append(current_pressure_gpa)
            except Exception as e:
                print(
                    "Failed to get energy or crystal structure "
                    f"at pressure {current_pressure_gpa} GPa. Exception:\n{repr(e)}"
                )
                # Don't keep trying larger pressures after one failure
                break
        return (
            binding_potential_energy_per_atom,
            binding_potential_energy_per_formula,
            volume_per_atom,
            volume_per_formula,
            density_kg_m3,
            pressure_gpa,
            a,
            parameter_values,
        )

    def _plot(
        self,
        ydata: List[float],
        title: str,
        ylabel: str,
        filename: str,
        keyname: str,
        y_logthresh: Optional[float] = None,
    ):
        min_press = min(self.pressure_gpa)
        max_press = max(self.pressure_gpa)
        min_press_range = min(abs(min_press), abs(max_press))
        fig, ax = plt.subplots()
        ax.plot(self.pressure_gpa, ydata, "-", color="k")
        fig.suptitle(title)
        # If pressure range is highly asymmetric, do symlog scale
        if max(abs(min_press), abs(max_press)) > min_press_range * 10:
            ax.set_xscale("symlog", linthresh=min_press_range)

            # Take care of major ticks
            ax.xaxis.get_major_locator().set_params(subs=[1, -1])
            major_ticks = list(ax.get_xticks())
            oldlabels = ax.get_xticklabels()
            newlabels = []
            closest_lin_decade = round(log(min_press_range, 10))
            del_label_locs = []
            if closest_lin_decade < 0:
                # just in case, delete 3 decades smaller. Won't hurt anything
                # Don't delete too many though, or you'll get np.allclose to zero
                irange = range(closest_lin_decade - 3, closest_lin_decade)
            else:
                irange = range(0, closest_lin_decade)
            for i in irange:
                del_label_locs += [-(10**i), 10**i]
            # Hide labels that are too crowded
            for tick, label in zip(major_ticks, oldlabels):
                if any([np.isclose(tick, del_loc) for del_loc in del_label_locs]):
                    newlabels.append("")
                else:
                    newlabels.append(label)
            ax.set_xticks(major_ticks, newlabels)

            # Take care of minor ticks
            ax.xaxis.get_minor_locator().set_params(
                subs=list(range(-9, 0)) + list(range(1, 10))
            )
            oldminorticks = list(ax.get_xticks(minor=True))
            newminorticks = []
            tickstep = 10 ** (closest_lin_decade - 1)
            for i in range(
                -floor(min_press_range / tickstep),
                floor(min_press_range / tickstep) + 1,
            ):
                if not any(
                    [np.isclose(i * tickstep, old_tick) for old_tick in oldminorticks]
                ):
                    newminorticks.append(i * tickstep)
            ax.set_xticks(oldminorticks + newminorticks, minor=True)
            ax.axvspan(min_press_range, max_press, color="b", alpha=0.18)
            pad_x = 20.0
            ax.text(
                0.95,
                -0.08,
                "(Log scale)",
                verticalalignment="top",
                horizontalalignment="right",
                transform=ax.transAxes,
                fontsize="small",
            )
            ax.text(
                0.05,
                -0.08,
                "(Linear scale)",
                verticalalignment="top",
                horizontalalignment="left",
                transform=ax.transAxes,
                fontsize="small",
            )
        else:
            pad_x = 4.0

        if y_logthresh is not None:
            # If somehow we get negative numbers that are bigger
            # than y_logthresh, we want to extend the linear region
            # to them so we don't leave the linear region at the
            # bottom of the graph
            y_logthresh = max(-min(ydata), y_logthresh)
            ax.set_yscale("symlog", linthresh=y_logthresh)

            # Take care of major ticks
            ax.yaxis.get_major_locator().set_params(subs=[1, -1])
            major_ticks = list(ax.get_yticks())
            oldlabels = ax.get_yticklabels()
            newlabels = []
            closest_lin_decade = round(log(y_logthresh, 10))
            del_label_locs = []
            if closest_lin_decade < 0:
                # just in case, delete 3 decades smaller. Won't hurt anything
                # Don't delete too many though, or you'll get np.allclose to zero
                irange = range(closest_lin_decade - 3, closest_lin_decade)
            else:
                irange = range(0, closest_lin_decade)
            for i in irange:
                del_label_locs += [-(10**i), 10**i]
            # Hide labels that are too crowded
            for tick, label in zip(major_ticks, oldlabels):
                if any([np.isclose(tick, del_loc) for del_loc in del_label_locs]):
                    newlabels.append("")
                else:
                    newlabels.append(label)
            ax.set_yticks(major_ticks, newlabels)

            # Take care of minor ticks
            ax.yaxis.get_minor_locator().set_params(
                subs=list(range(-9, 0)) + list(range(1, 10))
            )
            oldminorticks = list(ax.get_yticks(minor=True))
            newminorticks = []
            tickstep = 10 ** (closest_lin_decade - 1)
            for i in range(
                -floor(y_logthresh / tickstep),
                floor(y_logthresh / tickstep) + 1,
            ):
                if not any(
                    [np.isclose(i * tickstep, old_tick) for old_tick in oldminorticks]
                ):
                    newminorticks.append(i * tickstep)
            ax.set_yticks(oldminorticks + newminorticks, minor=True)
            ax.axhspan(min(ydata), y_logthresh, color="g", alpha=0.18)
            pad_y = 24.0
            ax.text(
                -0.1,
                0.95,
                "(Log scale)",
                verticalalignment="top",
                horizontalalignment="right",
                transform=ax.transAxes,
                fontsize="small",
                rotation="vertical",
            )
            ax.text(
                -0.1,
                0.05,
                "(Linear scale)",
                verticalalignment="bottom",
                horizontalalignment="right",
                transform=ax.transAxes,
                fontsize="small",
                rotation="vertical",
            )
            ax.set_ylim(min(ydata), max(ydata))
        else:
            # If we have an all-linear scale, and including
            # zero in the won't squish the data too much
            # (by less than half), then include zero
            if min(ydata) > 0 and min(ydata) < max(ydata) / 2:
                ax.set_ylim(0, max(ydata))
            else:
                ax.set_ylim(min(ydata), max(ydata))
            pad_y = 4.0
        ax.set_xlim(min_press, max_press)
        ax.set_xlabel("Pressure (GPa)", labelpad=pad_x)
        ax.set_ylabel(ylabel, labelpad=pad_y)
        plt.savefig(filename, bbox_inches="tight")
        self._add_file_to_current_property_instance(keyname, filename)

    def _calculate(
        self,
        pressure_step_gpa: float = 0.5,
        min_fractional_volume: float = 0.25,
        max_fractional_volume: float = 4,
        max_pressure_steps: int = 1000,
        exp_step: float = log(2) / 6,
        min_fmax: float = 1e-5,
        min_steps: int = 200,
        min_algorithm: type[Optimizer] = LBFGSLineSearch,
        cell_filter: type[UnitCellFilter] = FrechetCellFilter,
        min_opt_kwargs: Dict = {},
        min_flt_kwargs: Dict = {},
        **kwargs,
    ) -> None:
        """
        Computes the energy and (generally anisotropic) crystal structure as a function
        of applied hydrostatic pressure. Note that reaching either maximum or minimum
        requested fractional volume is not guaranteed, as the computation may fail
        before then (e.g. due to neighbor list overflow). All arguments are optional,
        and have reasonable defaults that should work for most materials.

        Args:
            pressure_step_gpa:
                pressure step
            min_fractional_volume:
                Stop the pressure scan if the system gets below this volume
            max_fractional_volume:
                Stop the pressure scan if the system gets above this volume
            max_pressure_steps:
                A hard limit on the maximum steps if the volume limits are
                not reached somehow
            exp_step:
                The compression scan grows exponentially,
                p=exp(i*exp_step)*pressure_step_gpa.
            min_fmax:
                The force tolerance used for minimization
            min_steps:
                The step limit for minimization
            min_algorithm:
                The ASE Optimizer class type to use for minimization
            cell_filter:
                The ASE UnitCellFilter class type to use for minimization
            min_opt_kwargs:
                A dictionary of keyword arguments to pass to the optimizer
            min_flt_kwargs:
                A dictionary of keyword arguments to pass to the filter
                in addition to the scalar_pressure argument
        """
        if min_fractional_volume > 1.0:
            raise RuntimeError(
                "Minimum fractional volume requested must be less than 1"
            )
        if max_fractional_volume < 1.0:
            raise RuntimeError(
                "Maximum fractional volume requested must be greater than 1"
            )
        self.pressure_step_gpa = pressure_step_gpa
        self.original_atoms = self._get_atoms()
        original_volume = self.original_atoms.get_volume()
        self.min_volume = original_volume * min_fractional_volume
        self.max_volume = original_volume * max_fractional_volume
        self.max_pressure_steps = max_pressure_steps
        self.exp_step = exp_step
        self.min_fmax = min_fmax
        self.min_steps = min_steps
        self.min_algorithm = min_algorithm
        self.cell_filter = cell_filter
        self.min_opt_kwargs = min_opt_kwargs
        self.min_flt_kwargs = min_flt_kwargs

        self.disclaimer = None

        print("\nPerforming energy scan...\n")

        # Forward scan
        (
            binding_potential_energy_per_atom,
            binding_potential_energy_per_formula,
            volume_per_atom,
            volume_per_formula,
            density_kg_m3,
            self.pressure_gpa,
            a,
            parameter_values,
        ) = self._do_scan(backward=False)

        max_pe_peratom_compr = max(binding_potential_energy_per_atom)

        # Backward scan
        (
            binding_potential_energy_per_atom_back,
            binding_potential_energy_per_formula_back,
            volume_per_atom_back,
            volume_per_formula_back,
            density_kg_m3_back,
            pressure_gpa_back,
            a_back,
            parameter_values_back,
        ) = self._do_scan(backward=True)

        max_pe_peratom_vacu = max(binding_potential_energy_per_atom_back)
        pe_peratom_eqbm = binding_potential_energy_per_atom_back[0]
        rel_max_pe_peratom_compr = max_pe_peratom_compr - pe_peratom_eqbm
        rel_max_pe_peratom_vacu = max_pe_peratom_vacu - pe_peratom_eqbm

        binding_potential_energy_per_atom = (
            list(reversed(binding_potential_energy_per_atom_back))
            + binding_potential_energy_per_atom
        )
        binding_potential_energy_per_formula = (
            list(reversed(binding_potential_energy_per_formula_back))
            + binding_potential_energy_per_formula
        )
        volume_per_atom = list(reversed(volume_per_atom_back)) + volume_per_atom
        volume_per_formula = (
            list(reversed(volume_per_formula_back)) + volume_per_formula
        )
        density_kg_m3 = list(reversed(density_kg_m3_back)) + density_kg_m3
        self.pressure_gpa = list(reversed(pressure_gpa_back)) + self.pressure_gpa
        a = list(reversed(a_back)) + a
        parameter_values = list(reversed(parameter_values_back)) + parameter_values

        assert len(binding_potential_energy_per_atom) == len(self.pressure_gpa)
        assert len(binding_potential_energy_per_formula) == len(self.pressure_gpa)
        assert len(volume_per_atom) == len(self.pressure_gpa)
        assert len(volume_per_formula) == len(self.pressure_gpa)
        assert len(density_kg_m3) == len(self.pressure_gpa)
        assert len(a) == len(self.pressure_gpa)
        assert len(parameter_values) == len(self.pressure_gpa)

        # Infer file names

        # Now it is time to write the output.
        parameter_names = self.get_nominal_parameter_names()
        if parameter_names is None:
            parameter_names = []

        self._add_property_instance_and_common_crystal_genome_keys(
            property_name=(
                "energy-and-crystal-structure-vs-hydrostatic-pressure-relation"
            ),
            write_stress=False,
            write_temp=False,
            disclaimer=self.disclaimer,
            omit_keys=[
                # Crystal structures are arrays due to the scan
                "a",
                "parameter-values",
                "coordinates-file",
                "coordinates-file-conventional",
            ],
        )
        self._add_key_to_current_property_instance(
            "volume-per-atom", volume_per_atom, unit="angstrom^3"
        )
        self._add_key_to_current_property_instance(
            "volume-per-formula", volume_per_formula, unit="angstrom^3"
        )
        self._add_key_to_current_property_instance(
            "mass-density", density_kg_m3, unit="kg/m^3"
        )
        self._add_key_to_current_property_instance(
            "binding-potential-energy-per-atom",
            binding_potential_energy_per_atom,
            unit="eV",
        )
        self._add_key_to_current_property_instance(
            "binding-potential-energy-per-formula",
            binding_potential_energy_per_formula,
            unit="eV",
        )
        self._add_key_to_current_property_instance("a", a, unit="angstrom")
        if len(parameter_names) != 0:
            self._add_key_to_current_property_instance(
                "parameter-values",
                parameter_values,
            )
        self._add_key_to_current_property_instance(
            "pressure", self.pressure_gpa, unit="GPa"
        )

        # 1-based, and we have already created the current prop instance
        prop_idx = len(self.property_instances)
        coordinates_file = []
        # Infer names of saved files
        for p_gpa in self.pressure_gpa:
            filename = CONFIG_TPL.format(p_gpa=p_gpa, prop_idx=prop_idx)
            assert os.path.isfile(os.path.join("output", filename))
            coordinates_file.append(filename)

        self._add_key_to_current_property_instance("coordinates-file", coordinates_file)
        # Plots

        # Energy
        # If one arm of the energy curve is much higher than the other, we should do
        # a symlog scale
        if rel_max_pe_peratom_compr > 10 * rel_max_pe_peratom_vacu:
            # Linear scale > 1 eV loses detail
            rel_pe_peratom_logthresh = min(rel_max_pe_peratom_vacu, 1)
        elif rel_max_pe_peratom_vacu > 10 * rel_max_pe_peratom_compr:
            # Linear scale > 1 eV loses detail
            rel_pe_peratom_logthresh = min(rel_max_pe_peratom_compr, 1)
        else:
            rel_pe_peratom_logthresh = None

        self._plot(
            ydata=np.asarray(binding_potential_energy_per_atom) - pe_peratom_eqbm,
            title=(
                "Binding potential energy per atom\n"
                "as a function of applied hydrostatic pressure"
            ),
            ylabel=f"Energy relative to {pe_peratom_eqbm:.3f} (eV/atom)",
            filename="eplot.svg",
            keyname="binding-potential-energy-per-atom-vs-pressure-curve",
            y_logthresh=rel_pe_peratom_logthresh,
        )

        # Volume
        self._plot(
            ydata=volume_per_atom,
            title="Volume per atom as a function of applied hydrostatic pressure",
            ylabel="Volume (angstrom\u00b3/atom)",
            filename="vplot.svg",
            keyname="volume-per-atom-vs-pressure-curve",
        )

        # Density
        self._plot(
            ydata=density_kg_m3,
            title="Mass density as a function of applied hydrostatic pressure",
            ylabel="Density (kg/m\u00b3)",
            filename="rhoplot.svg",
            keyname="density-vs-pressure-curve",
        )

        # a
        A_IT = r"$a$"
        self._plot(
            ydata=a,
            title=(
                f"Lattice parameter {A_IT} "
                "as a function of applied hydrostatic pressure"
            ),
            ylabel=f"{A_IT} (angstrom)",
            filename="aplot.svg",
            keyname="a-vs-pressure-curve",
        )
        # Other parameter_values
        # Convert parameter_values to numpy array so we can get slices
        parameter_values = np.asarray(parameter_values)

        B_IT = r"$b$"
        C_IT = r"$c$"

        tex_map = {
            "b/a": f"{B_IT}/{A_IT}",
            "c/a": f"{C_IT}/{A_IT}",
            "alpha": r"$\alpha$",
            "beta": r"$\beta$",
            "gamma": r"$\gamma$",
        }

        for i, param_name in enumerate(parameter_names):
            if (
                param_name.startswith("x")
                or param_name.startswith("y")
                or param_name.startswith("z")
            ):
                break  # not continue, if we are here, we are done
            param_valu_arr = parameter_values[:, i]
            nice_name = param_name.replace("/", "-over-")
            filename = f"{nice_name}plot.svg"
            keyname = f"{nice_name}-vs-pressure-curve"
            self._plot(
                ydata=param_valu_arr,
                title=(
                    f"Lattice parameter {tex_map[param_name]} as a function of "
                    "applied hydrostatic pressure"
                ),
                ylabel=tex_map[param_name],
                filename=filename,
                keyname=keyname,
            )



import logging
from dataclasses import dataclass
from kim_tools import SingleCrystalTestDriver
from kim_tools.test_driver.core import FMAX_INITIAL, MAXSTEPS_INITIAL
from kim_tools.symmetry_util.core import (
    fit_voigt_tensor_to_cell_and_space_group,
    fit_voigt_tensor_and_error_to_cell_and_space_group,
)
from kim_tools.symmetry_util.elasticity import (
    indep_elast_compon_names_and_values_from_voigt,
    calc_bulk,
    find_nearest_isotropy,
    map_to_Kelvin,
    voigt_elast_struct_svg,
)
from kim_tools.aflow_util.core import AFLOW_EXECUTABLE
from typing import Optional, Union, Dict
from numdifftools import MaxStepGenerator
from ase.units import GPa
from ase.optimize.lbfgs import LBFGSLineSearch
from ase.optimize.optimize import Optimizer
from ase.calculators.calculator import Calculator
from .elastic import ElasticConstants
import numpy as np
import numpy.typing as npt

logger = logging.getLogger(__name__)
logging.basicConfig(filename="kim-tools.log", level=logging.INFO, force=True)


@dataclass
class ElConstResultsErrors:
    """
    All information of interest from a given run of ElasticConstants.results()
    """

    elconst_raw: npt.ArrayLike
    elconst_raw_err: npt.ArrayLike
    elconst_raw_err_frac: float
    elconst_matl_symm: npt.ArrayLike
    elconst_matl_symm_err: npt.ArrayLike
    matl_symm_err_frac: float
    message: str


def compute_results_summary(
    elconst_raw: npt.ArrayLike,
    elconst_raw_err: npt.ArrayLike,
    message: str,
    cell: npt.ArrayLike,
    sgnum: Union[int, str],
) -> ElConstResultsErrors:
    """
    Make an ElConstResultsErrors object from the output of
    ``ElasticConstants.results()``
    """
    econst_raw_Kelv = map_to_Kelvin(elconst_raw)
    elconst_raw_norm = np.einsum("ij,ij", econst_raw_Kelv, econst_raw_Kelv)
    elconst_raw_norm = elconst_raw_norm**0.5

    addtl_msg = ""

    if np.isnan(elconst_raw_err).any():
        elconst_raw_err_frac = np.nan
        elconst_matl_symm = fit_voigt_tensor_to_cell_and_space_group(
            voigt_input=elconst_raw, cell=cell, sgnum=sgnum
        )
        elconst_matl_symm_err = np.full((6, 6), np.nan)
    else:
        elconst_raw_err_Kelv = map_to_Kelvin(elconst_raw_err)
        elconst_raw_err_norm = np.einsum(
            "ij,ij", elconst_raw_err_Kelv, elconst_raw_err_Kelv
        )
        elconst_raw_err_norm = elconst_raw_err_norm**0.5

        elconst_raw_err_frac = elconst_raw_err_norm / elconst_raw_norm

        elconst_matl_symm, elconst_matl_symm_err = (
            fit_voigt_tensor_and_error_to_cell_and_space_group(
                voigt_input=elconst_raw,
                voigt_error=elconst_raw_err,
                cell=cell,
                sgnum=sgnum,
                symmetric=True,
            )
        )
        addtl_msg += f"\nRelative norm of error estimate: {elconst_raw_err_frac}"

    matl_symm_err = elconst_matl_symm - elconst_raw

    matl_symm_err_Kelv = map_to_Kelvin(matl_symm_err)
    matl_symm_err_norm = np.einsum("ij,ij", matl_symm_err_Kelv, matl_symm_err_Kelv)
    matl_symm_err_norm = matl_symm_err_norm**0.5

    matl_symm_err_frac = matl_symm_err_norm / elconst_raw_norm

    addtl_msg += (
        f"\nRelative norm of deviation from material symmetry: {matl_symm_err_frac}"
    )

    print(addtl_msg)

    return ElConstResultsErrors(
        elconst_raw=elconst_raw,
        elconst_raw_err=elconst_raw_err,
        elconst_raw_err_frac=elconst_raw_err_frac,
        elconst_matl_symm=elconst_matl_symm,
        elconst_matl_symm_err=elconst_matl_symm_err,
        matl_symm_err_frac=matl_symm_err_frac,
        message=message + addtl_msg,
    )


def makeDefaultStepGenerator(base_step: float) -> MaxStepGenerator:
    return MaxStepGenerator(
        base_step=base_step,
        num_steps=14,
        use_exact_steps=True,
        step_ratio=1.6,
        offset=0,
    )


class TestDriver(SingleCrystalTestDriver):
    def __init__(
        self,
        model: Union[str, Calculator],
        suppr_sm_lmp_log: bool = True,
        aflow_executable: str = AFLOW_EXECUTABLE,
    ) -> None:
        """
        Args:
            model:
                ASE calculator or KIM model name to use
            suppr_sm_lmp_log:
                Suppress writing a lammps.log
            aflow_executable:
                Path to AFLOW executable
        """
        self.aflow_executable = aflow_executable
        super().__init__(model, suppr_sm_lmp_log=suppr_sm_lmp_log)

    def _calculate(
        self,
        method: Optional[str] = None,
        strain_step: Optional[Union[MaxStepGenerator, float]] = None,
        fmax: float = FMAX_INITIAL,
        steps: int = MAXSTEPS_INITIAL,
        algorithm: Optimizer = LBFGSLineSearch,
        fix_symmetry: bool = True,
        opt_kwargs: Dict = {},
        extrap_tol_95: float = 0.02,
        matl_symm_tol: float = 0.01,
        **kwargs,
    ) -> None:
        """
        Calculate the elasticity tensor (both raw and symmetrized according to material
        symmetry), unique elastic constant components, and distance to nearest isotropic
        elasticity tensor. By default, this Test Driver performs a loop over several
        settings to find the lowest error in a weighed combination of extrapolation
        error and deviation from material symmetry. If
        ``method`` and/or ``strain_step`` are provided, this behavior is overridden, and a
        single calculation is performed instead.

        Args:
            method:
                Select method for computing the elastic constants. The following
                methods are supported:
                "energy-condensed" : Compute elastic constants from the Hessian
                    of the condensed strain energy density (i.e. the enregy for
                    a given strain is relaxed with respect to internal atom
                    positions)
                "stress-condensed" : Compute elastic constants from the Jacobian
                    of the condensed stress (i.e. the stress for a given strain
                    where the energy is relaxed with respect to internal atom
                    positions)
                "energy-full" : Compute elastic constants from the full Hessian
                    relative to both strains and internal atom degrees of
                    freedom. This is followed by an algebraic manipulation to
                    account for the effect of atom relaxation on elastic
                    constants.
                In general, "energy-condensed" is the preferred method.
                The "stress-condensed" method is much faster, but generally less
                accurate. The "energy-full" method has accuracy comparable
                to "energy-condensed" but tends to be much slower due to the
                larger Hessian matrix that has to be computed.
                If `strain_step` is provided, but this is not, this will be set
                to "energy-condensed".
            strain_step:
                The step or ``numdifftools.MaxStepGenerator`` to use for numerical
                differentiation.
                If ``method`` is provided, but this is not, this will be set
                to:
                    MaxStepGenerator(
                        base_step=1e-4,
                        num_steps=14,
                        use_exact_steps=True,
                        step_ratio=1.6,
                        offset=0,
                    )
            fmax:
                Force convergence tolerance (the magnitude of the force on each
                atom must be less than this for convergence)
            steps:
                Maximum number of iterations for the minimization
            algorithm:
                ASE optimizer algorithm
            fix_symmetry:
                Whether to fix the symmetry during minimization
            opt_kwargs:
                Dictionary of kwargs to pass to optimizer
            extrap_tol_95:
                If neither ``method`` nor ``strain_step`` are specified,
                this is the tolerance that will be used to threshold
                and weight the 95% extrap error reported by
                numdifftools. The norm of the error matrix in Kelvin
                form divided by the norm of the elasticity matrix in
                Kelvin form will be compared to this tolerance.
            matl_symm_tol:
                If neither ``method`` nor ``strain_step`` are specified,
                this is the tolerance that will be used to threshold
                and weight the deviation from material symmetry.
                The deviation is the difference between the elastic constants
                symmetrized to fit the space group of the material and
                the as-computed elastic constants.
                The norm of the deviation matrix in Kelvin
                form divided by the norm of the elasticity matrix in
                Kelvin form will be compared to this tolerance.
        """

        print("\nE L A S T I C  C O N S T A N T  C A L C U L A T I O N S\n")
        print()

        prototype_label = self._get_nominal_crystal_structure_npt()["prototype-label"][
            "source-value"
        ]
        sgnum = int(prototype_label.split("_")[2])

        if fix_symmetry:
            sgnum_for_min = sgnum
        else:
            sgnum_for_min = 1

        atoms = self._get_atoms()

        o_cell = atoms.get_cell()

        moduli = ElasticConstants(
            atoms=atoms,
            sgnum=sgnum_for_min,
            fmax=fmax,
            steps=steps,
            algorithm=algorithm,
            opt_kwargs=opt_kwargs,
        )

        extrap_err_undef = False
        # Set up sequence of calculations
        if method is not None or strain_step is not None:
            if method is None:
                method = "energy-condensed"
            if strain_step is None:
                strain_step = makeDefaultStepGenerator(1e-4)
            else:
                if isinstance(strain_step, float):
                    # Can't get extrapolation error with a single step
                    extrap_err_undef = True
            sequence = [(method, strain_step)]
        else:
            sequence = []
            # From testing, only one model out of 388
            # benefitted from having a strain step of 1e-2
            # with energy-condensed
            for base_step in [1e-4, 1e-3]:
                sequence.append(
                    ("energy-condensed", makeDefaultStepGenerator(base_step))
                )
            for base_step in [1e-4, 1e-3, 1e-2, 0.1]:
                sequence.append(
                    ("stress-condensed", makeDefaultStepGenerator(base_step))
                )

        results_summaries = []
        converged = False
        for method, step in sequence:
            try:
                results = moduli.results(
                    method=method,
                    step=step,
                )
                results_summaries.append(
                    compute_results_summary(*results, o_cell, sgnum)
                )
                if (results_summaries[-1].elconst_raw_err_frac < extrap_tol_95) and (
                    results_summaries[-1].matl_symm_err_frac < matl_symm_tol
                ):
                    # If elconst_raw_err_frac is np.nan, will never get here
                    converged = True
                    break
            except Exception as e:
                msg = (
                    "\nThe following exception was caught during "
                    "Hessian or Jacobian calculation:\n"
                )
                msg += repr(e)
                print(msg)
                logger.info(msg)

        if results_summaries == []:
            raise RuntimeError(
                "All attempts to compute elastic constants failed. See log for error(s)"
            )
        # Check that presence or absence of NaNs makes sense given what we ran
        if any([np.isnan(result.elconst_raw_err_frac) for result in results_summaries]):
            assert extrap_err_undef, (
                "Unexpectedly got a NaN extrapolation error. This should only happen "
                "if a user asked for a single strain step"
            )
        # Check the converse -- if the user asked for a single strain step, we should
        # not have any extrapolation error estimates, and we should only have one result
        if extrap_err_undef:
            assert (
                len(results_summaries) == 1
            ), "We should never have more than one result if the user specified a step"
            assert np.isnan(results_summaries[0].elconst_raw_err_frac), (
                "If the user specified a single strain step, the extrapolation error "
                "must be nan"
            )

        if converged:
            disclaimer = None
        else:
            if extrap_err_undef:
                disclaimer = (
                    "Elastic constants were evaluated with a single step. Unable to "
                    "compute uncertainty."
                )
            else:
                disclaimer = (
                    r"Elastic constants calculation had a relative 95% uncertainty "
                    f"greater than {extrap_tol_95} and/or relative deviation from "
                    f"material symmetry greater than {matl_symm_tol}.\n"
                    "See stdout and logs for calculation details."
                )
            print(disclaimer)
            if len(results_summaries) > 1:
                # We did a loop but didn't converge. Need to find the best result
                results_summaries.sort(
                    key=lambda x: (x.elconst_raw_err_frac / extrap_tol_95) ** 2
                    + (x.matl_symm_err_frac / matl_symm_tol) ** 2,
                    reverse=True,
                )
                print()
                print("The following run was chosen as having the lowest error: ")
                print(results_summaries[-1].message)

        # In all cases, the best result ended up at the end of the list. Assign
        # variables and convert. NaN will stay NaN, no special treatment
        # needed
        elconst_raw = results_summaries[-1].elconst_raw / GPa
        elconst_raw_err = results_summaries[-1].elconst_raw_err / GPa
        elconst_matl_symm = results_summaries[-1].elconst_matl_symm / GPa
        elconst_matl_symm_err = results_summaries[-1].elconst_matl_symm_err / GPa
        units = "GPa"

        elconst_names, elconst_values = indep_elast_compon_names_and_values_from_voigt(
            voigt=elconst_matl_symm, sgnum=sgnum
        )
        _, elconst_values_err = indep_elast_compon_names_and_values_from_voigt(
            voigt=elconst_matl_symm_err, sgnum=sgnum
        )

        bulk = calc_bulk(elconst_raw)
        # Compute nearest isotropic constants and distance
        try:
            d_iso, bulk_iso, shear_iso = find_nearest_isotropy(elconst_raw)
            got_iso = True
        except Exception:
            got_iso = False  # Failure can occur if elastic constants are
            # not positive definite

        # Echo output
        print("\nR E S U L T S\n")
        print(f"Elastic constants [{units}]:")
        print(
            np.array_str(
                elconst_raw,
                precision=5,
                max_line_width=100,
                suppress_small=True,
            )
        )
        if not extrap_err_undef:
            print()
            print(f"95 %% Error estimate [{units}]:")
            print(
                np.array_str(
                    elconst_raw_err,
                    precision=5,
                    max_line_width=100,
                    suppress_small=True,
                )
            )
        print()
        print(f"Bulk modulus [{units}] = {bulk}")
        print()
        print("Unique elastic constants for space group {} [{}]".format(sgnum, units))
        print(elconst_names)
        print(elconst_values)
        print()
        if got_iso:
            print("Nearest matrix of isotropic elastic constants:")
            print(f"Distance to isotropic state [-]  = {d_iso}")
            print(f"Isotropic bulk modulus      [{units}] = {bulk_iso}")
            print(f"Isotropic shear modulus     [{units}] = {shear_iso}")
        else:
            print("WARNING: Nearest isotropic state not computed.")

        ####################################################
        # ACTUAL CALCULATION ENDS
        ####################################################

        ####################################################
        # PROPERTY WRITING
        ####################################################
        self._add_property_instance_and_common_crystal_genome_keys(
            "elastic-constants-isothermal-npt",
            write_stress=True,
            write_temp=True,
            disclaimer=disclaimer,
        )
        self._add_key_to_current_property_instance(
            "elastic-constants-names", elconst_names
        )
        if not extrap_err_undef:
            self._add_key_to_current_property_instance(
                "elastic-constants-values",
                elconst_values,
                "GPa",
                {
                    "source-expand-uncert-value": elconst_values_err,
                    "uncert-lev-of-confid": 95,
                },
            )
            self._add_key_to_current_property_instance(
                "elasticity-matrix-raw",
                elconst_raw,
                "GPa",
                {
                    "source-expand-uncert-value": elconst_raw_err,
                    "uncert-lev-of-confid": 95,
                },
            )
            self._add_key_to_current_property_instance(
                "elasticity-matrix",
                elconst_matl_symm,
                "GPa",
                {
                    "source-expand-uncert-value": elconst_matl_symm_err,
                    "uncert-lev-of-confid": 95,
                },
            )
        else:
            self._add_key_to_current_property_instance(
                "elastic-constants-values",
                elconst_values,
                "GPa",
            )
            self._add_key_to_current_property_instance(
                "elasticity-matrix-raw",
                elconst_raw,
                "GPa",
            )
            self._add_key_to_current_property_instance(
                "elasticity-matrix",
                elconst_matl_symm,
                "GPa",
            )

        matrix_structure_file = "matrix-structure.svg"
        voigt_elast_struct_svg(sgnum, matrix_structure_file)
        self._add_file_to_current_property_instance(
            "elasticity-matrix-structure", matrix_structure_file
        )

        if got_iso:
            self._add_key_to_current_property_instance("distance-to-isotropy", d_iso)

        self._add_property_instance_and_common_crystal_genome_keys(
            "bulk-modulus-isothermal-npt",
            write_stress=True,
            write_temp=True,
            disclaimer=disclaimer,
        )
        self._add_key_to_current_property_instance(
            "isothermal-bulk-modulus", bulk, "GPa"
        )
